# -*- coding: utf-8 -*-
"""
Routines for handling of Reftek130 packets.

Currently only event header (EH), event trailer (ET) and data (DT) packets are
handled. These three packets have more or less the same meaning in the first 8
bytes of the payload which makes the first 24 bytes the so called extended
header.
"""
import sys
import warnings

import numpy as np

from obspy import UTCDateTime
from obspy.core.compatibility import from_buffer
from obspy.io.mseed.headers import clibmseed

from .util import (
    _decode_ascii, _parse_long_time, _16_tuple_ascii, _16_tuple_int,
    _16_tuple_float, bcd, bcd_hex,
    bcd_julian_day_string_to_nanoseconds_of_year, bcd_16bit_int, bcd_8bit_hex,
    _get_nanoseconds_for_start_of_year)


class Reftek130UnpackPacketError(ValueError):
    pass


PACKET_TYPES_IMPLEMENTED = ("EH", "ET", "DT")
PACKET_TYPES_NOT_IMPLEMENTED = ("AD", "CD", "DS", "FD", "OM", "SC", "SH")
PACKET_TYPES = PACKET_TYPES_IMPLEMENTED + PACKET_TYPES_NOT_IMPLEMENTED


# The extended header which is the same for EH/ET/DT packets.
# tuples are:
#  - field name
#  - dtype during initial reading
#  - conversion routine (if any)
#  - dtype after conversion
PACKET = [
    ("packet_type", "|S2", None, "S2"),
    ("experiment_number", np.uint8, bcd, np.uint8),
    ("year", np.uint8, bcd, np.uint8),
    ("unit_id", (np.uint8, 2), bcd_hex, "S4"),
    ("time", (np.uint8, 6), bcd_julian_day_string_to_nanoseconds_of_year,
     np.int64),
    ("byte_count", (np.uint8, 2), bcd_16bit_int, np.uint16),
    ("packet_sequence", (np.uint8, 2), bcd_16bit_int, np.uint16),
    ("event_number", (np.uint8, 2), bcd_16bit_int, np.uint16),
    ("data_stream_number", np.uint8, bcd, np.uint8),
    ("channel_number", np.uint8, bcd, np.uint8),
    ("number_of_samples", (np.uint8, 2), bcd_16bit_int, np.uint32),
    ("flags", np.uint8, None, np.uint8),
    ("data_format", np.uint8, bcd_8bit_hex, "S2"),
    # Temporarily store the payload here.
    ("payload", (np.uint8, 1000), None, (np.uint8, 1000))]


# name, offset, length (bytes) and converter routine for EH/ET packet payload
EH_PAYLOAD = {
    "trigger_time_message": (0, 33, _decode_ascii),
    "time_source": (33, 1, _decode_ascii),
    "time_quality": (34, 1, _decode_ascii),
    "station_name_extension": (35, 1, _decode_ascii),
    "station_name": (36, 4, _decode_ascii),
    "stream_name": (40, 16, _decode_ascii),
    "_reserved_2": (56, 8, _decode_ascii),
    "sampling_rate": (64, 4, float),
    "trigger_type": (68, 4, _decode_ascii),
    "trigger_time": (72, 16, _parse_long_time),
    "first_sample_time": (88, 16, _parse_long_time),
    "detrigger_time": (104, 16, _parse_long_time),
    "last_sample_time": (120, 16, _parse_long_time),
    "channel_adjusted_nominal_bit_weights": (136, 128, _16_tuple_ascii),
    "channel_true_bit_weights": (264, 128, _16_tuple_ascii),
    "channel_gain_code": (392, 16, _16_tuple_ascii),
    "channel_ad_resolution_code": (408, 16, _16_tuple_ascii),
    "channel_fsa_code": (424, 16, _16_tuple_ascii),
    "channel_code": (440, 64, _16_tuple_ascii),
    "channel_sensor_fsa_code": (504, 16, _16_tuple_ascii),
    "channel_sensor_vpu": (520, 96, _16_tuple_float),
    "channel_sensor_units_code": (616, 16, _16_tuple_ascii),
    "station_channel_number": (632, 48, _16_tuple_int),
    "_reserved_3": (680, 156, _decode_ascii),
    "total_installed_channels": (836, 2, int),
    "station_comment": (838, 40, _decode_ascii),
    "digital_filter_list": (878, 16, _decode_ascii),
    "position": (894, 26, _decode_ascii),
    "reftek_120": (920, 80, None)}


# mseed steim compression is big endian
if sys.byteorder == 'little':
    SWAPFLAG = 1
else:
    SWAPFLAG = 0


class Packet(object):
    _headers = ('experiment_number', 'unit_id', 'byte_count',
                'packet_sequence', 'time')

    @staticmethod
    def from_data(data):
        packet_type = data['packet_type'].decode("ASCII", "ignore")
        if packet_type in ("EH", "ET"):
            return EHPacket(data)
        elif packet_type == "DT":
            return DTPacket(data)
        else:
            msg = "Can not create Reftek packet for packet type '{}'"
            raise NotImplementedError(msg.format(packet_type))

    def __init__(self, data):
        raise NotImplementedError()

    @property
    def type(self):
        return self._data['packet_type'].item()

    def __getattr__(self, name):
        if name in self._headers:
            return self._data[name].item()

    @property
    def nanoseconds(self):
        return self._data['time'].item()

    @property
    def time(self):
        return UTCDateTime(ns=self._data['time'].item())


class EHPacket(Packet):
    __slots__ = ["_data"] + list(EH_PAYLOAD.keys())
    _headers = ('packet_sequence', 'experiment_number', 'unit_id',
                'byte_count', 'time', 'event_number', 'data_stream_number',
                'data_format', 'flags')

    def __init__(self, data):
        self._data = data
        payload = self._data["payload"].tobytes()
        for name, (start, length, converter) in EH_PAYLOAD.items():
            data = payload[start:start + length]
            if converter is not None:
                data = converter(data)
            setattr(self, name, data)

    def _to_dict(self):
        """
        Convert to dictionary structure.
        """
        return {key: getattr(self, key) for key in EH_PAYLOAD.keys()}

    def __str__(self, compact=False):
        if compact:
            sta = (self.station_name.strip() +
                   self.station_name_extension.strip())
            info = ("{:04d} {:2s} {:4s} {:2d} {:4d} {:4d} {:2d} {:2s} "
                    "{:5s}  {:4s}        {!s}").format(
                        self.packet_sequence, self.type.decode(),
                        self.unit_id.decode(), self.experiment_number,
                        self.byte_count, self.event_number,
                        self.data_stream_number, self.data_format.decode(),
                        sta, str(self.sampling_rate)[:4], self.time)
        else:
            info = []
            for key in self._headers:
                value = getattr(self, key)
                if key in ("unit_id", "data_format"):
                    value = value.decode()
                info.append("{}: {}".format(key, value))
            info.append("-" * 20)
            for key in sorted(EH_PAYLOAD.keys()):
                value = getattr(self, key)
                if key in ("trigger_time", "detrigger_time",
                           "first_sample_time", "last_sample_time"):
                    if value is not None:
                        value = UTCDateTime(ns=value)
                info.append("{}: {}".format(key, value))
            info = "{} Packet\n\t{}".format(self.type.decode(),
                                            "\n\t".join(info))
        return info


class DTPacket(Packet):
    __slots__ = ["_data"]
    _headers = ('packet_sequence', 'experiment_number', 'unit_id',
                'byte_count', 'time', 'event_number', 'data_stream_number',
                'channel_number', 'number_of_samples', 'data_format', 'flags')

    def __init__(self, data):
        self._data = data

    def __str__(self, compact=False):
        if compact:
            info = ("{:04d} {:2s} {:4s} {:2d} {:4d} {:4d} {:2d} {:2s} "
                    "           {:2d} {:4d} {!s}").format(
                        self.packet_sequence, self.type.decode(),
                        self.unit_id.decode(), self.experiment_number,
                        self.byte_count, self.event_number,
                        self.data_stream_number, self.data_format.decode(),
                        self.channel_number, self.number_of_samples, self.time)
        else:
            info = []
            for key in self._headers:
                value = getattr(self, key)
                if key in ("unit_id", "data_format"):
                    value = value.decode()
                info.append("{}: {}".format(key, value))
            info = "{} Packet\n\t{}".format(self.type.decode(),
                                            "\n\t".join(info))
        return info


PACKET_INITIAL_UNPACK_DTYPE = np.dtype([
    (name, dtype_initial)
    for name, dtype_initial, converter, dtype_final in PACKET])

PACKET_FINAL_DTYPE = np.dtype([
    (name, dtype_final)
    for name, dtype_initial, converter, dtype_final in PACKET])


def _initial_unpack_packets(bytestring):
    """
    First unpack data with dtype matching itemsize of storage in the reftek
    file, than allocate result array with dtypes for storage of python
    objects/arrays and fill it with the unpacked data.
    """
    if not len(bytestring):
        return np.array([], dtype=PACKET_FINAL_DTYPE)

    if len(bytestring) % 1024 != 0:
        tail = len(bytestring) % 1024
        bytestring = bytestring[:-tail]
        msg = ("Length of data not a multiple of 1024. Data might be "
               "truncated. Dropping {:d} byte(s) at the end.").format(tail)
        warnings.warn(msg)
    data = from_buffer(
        bytestring, dtype=PACKET_INITIAL_UNPACK_DTYPE)
    result = np.empty_like(data, dtype=PACKET_FINAL_DTYPE)

    for name, dtype_initial, converter, dtype_final in PACKET:
        if converter is None:
            result[name][:] = data[name][:]
        else:
            try:
                result[name][:] = converter(data[name])
            except Exception as e:
                raise Reftek130UnpackPacketError(str(e))
    # time unpacking is special and needs some additional work.
    # we need to add the POSIX timestamp of the start of respective year to the
    # already unpacked seconds into the respective year..
    result['time'][:] += [_get_nanoseconds_for_start_of_year(y)
                          for y in result['year']]
    return result


def _unpack_C0_C2_data(packets, encoding):  # noqa
    """
    Unpacks sample data from a packet array that uses 'C0' or 'C2' data
    encoding.

    :type packets: :class:`numpy.ndarray` (dtype ``PACKET_FINAL_DTYPE``)
    :param packets: Array of data packets (``packet_type`` ``'DT'``) from which
        to unpack the sample data (with data encoding 'C0' or 'C2').
    :type encoding: str
    :param encoding: Reftek data encoding as specified in event header (EH)
        packet, either ``'C0'`` or ``'C2'``.
    """
    if encoding == 'C0':
        encoding_bytes = b'C0'
    elif encoding == 'C2':
        encoding_bytes = b'C2'
    else:
        msg = "Unregonized encoding: '{}'".format(encoding)
        raise ValueError(msg)
    if np.any(packets['data_format'] != encoding_bytes):
        differing_formats = np.unique(
            packets[packets['data_format'] !=
                    encoding_bytes]['data_format']).tolist()
        msg = ("Using '{}' data format unpacking routine but some packet(s) "
               "specify other data format(s): {}".format(encoding,
                                                         differing_formats))
        warnings.warn(msg)
    # if the packet array is contiguous in memory (which it generally should
    # be), we can work with the memory address of the first packed data byte
    # and advance it by a fixed offset when moving from one packet to the next
    if packets.flags['C_CONTIGUOUS'] and packets.flags['F_CONTIGUOUS']:
        return _unpack_C0_C2_data_fast(packets, encoding)
    # if the packet array is *not* contiguous in memory, fall back to slightly
    # slower unpacking with looking up the memory position of the first packed
    # byte in each packet individually.
    else:
        return _unpack_C0_C2_data_safe(packets, encoding)


def _unpack_C0_C2_data_fast(packets, encoding):  # noqa
    """
    Unpacks sample data from a packet array that uses 'C0' or 'C2' data
    encoding.

    Unfortunately the whole data cannot be unpacked with one call to
    libmseed as some payloads do not take the full 960 bytes. They are
    thus padded which would results in padded pieces directly in a large
    array and libmseed (understandably) does not support that.

    Thus we resort to *tada* pointer arithmetics in Python ;-) This is
    quite a bit faster then correctly casting to an integer pointer so
    it's worth it.

    Also avoid a data copy.

    Writing this directly in C would be about 3 times as fast so it might
    be worth it.

    :type packets: :class:`numpy.ndarray` (dtype ``PACKET_FINAL_DTYPE``)
    :param packets: Array of data packets (``packet_type`` ``'DT'``) from which
        to unpack the sample data (with data encoding 'C0' or 'C2').
    :type encoding: str
    :param encoding: Reftek data encoding as specified in event header (EH)
        packet, either ``'C0'`` or ``'C2'``.
    """
    if encoding == 'C0':
        decode_steim = clibmseed.msr_decode_steim1
    elif encoding == 'C2':
        decode_steim = clibmseed.msr_decode_steim2
    else:
        msg = "Unregonized encoding: '{}'".format(encoding)
        raise ValueError(msg)
    npts = packets["number_of_samples"].sum()
    unpacked_data = np.empty(npts, dtype=np.int32)
    pos = 0
    s = packets[0]["payload"][40:].ctypes.data
    if len(packets) > 1:
        offset = (
            packets[1]["payload"][40:].ctypes.data - s)
    else:
        offset = 0
    for _npts in packets["number_of_samples"]:
        decode_steim(
            s, 960, _npts, unpacked_data[pos:], _npts, None,
            SWAPFLAG)
        pos += _npts
        s += offset
    return unpacked_data


def _unpack_C0_C2_data_safe(packets, encoding):  # noqa
    """
    Unpacks sample data from a packet array that uses 'C0' or 'C2' data
    encoding.

    If the packet array is *not* contiguous in memory, fall back to slightly
    slower unpacking with looking up the memory position of the first packed
    byte in each packet individually.

    :type packets: :class:`numpy.ndarray` (dtype ``PACKET_FINAL_DTYPE``)
    :param packets: Array of data packets (``packet_type`` ``'DT'``) from which
        to unpack the sample data (with data encoding 'C0' or 'C2').
    :type encoding: str
    :param encoding: Reftek data encoding as specified in event header (EH)
        packet, either ``'C0'`` or ``'C2'``.
    """
    if encoding == 'C0':
        decode_steim = clibmseed.msr_decode_steim1
    elif encoding == 'C2':
        decode_steim = clibmseed.msr_decode_steim2
    else:
        msg = "Unregonized encoding: '{}'".format(encoding)
        raise ValueError(msg)
    npts = packets["number_of_samples"].sum()
    unpacked_data = np.empty(npts, dtype=np.int32)
    pos = 0
    # Sample data starts at byte 40 in the DT packet's
    # payload.
    for p in packets:
        _npts = p["number_of_samples"]
        decode_steim(
            p["payload"][40:].ctypes.data, 960, _npts,
            unpacked_data[pos:], _npts, None, SWAPFLAG)
        pos += _npts
    return unpacked_data


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



# -*- coding: utf-8 -*-
import os
import re
import shutil
import tempfile

import numpy as np

from obspy import UTCDateTime, Trace, Stream
from obspy.core.util.misc import TemporaryWorkingDirectory
from obspy.clients.filesystem.sds import SDS_FMTSTR, Client
from obspy.scripts.sds_html_report import main as sds_report


class TemporarySDSDirectory(object):
    """
    Handles creation and deletion of a temporary SDS directory structure.
    To be used with "with" statement.
    """
    sampling_rate = 0.1
    networks = ("AB", "CD")
    stations = ("XYZ", "ZZZ3")
    locations = ("", "00")
    channels = ("HHZ", "HHN", "HHE", "BHZ", "BHN", "BHE")

    def __init__(self, year, doy, time=None):
        """
        Set which day's midnight (00:00 hours) is used as a day break in the
        testing (to split the test data into two files).

        If `time` is specified it overrides `year` and `doy`.
        """
        if time:
            self.time = time
        else:
            self.time = UTCDateTime("%d-%03dT00:00:00" % (year, doy))
        delta = 1.0 / self.sampling_rate

        self.stream = Stream()
        for net in self.networks:
            for sta in self.stations:
                for loc in self.locations:
                    for cha in self.channels:
                        tr = Trace(
                            data=np.arange(100, dtype=np.int32),
                            header=dict(
                                network=net, station=sta, location=loc,
                                channel=cha, sampling_rate=self.sampling_rate,
                                starttime=self.time - 30 * delta))

                        # cut into two seamless traces
                        tr1 = tr.slice(endtime=self.time + 5 * delta)
                        tr2 = tr.slice(starttime=self.time + 6 * delta)
                        self.stream.append(tr1)
                        self.stream.append(tr2)

    def __enter__(self):
        self.old_dir = os.getcwd()
        self.tempdir = tempfile.mkdtemp(prefix='obspy-sdstest-')
        for tr_ in self.stream:
            t_ = tr_.stats.starttime
            full_path = SDS_FMTSTR.format(year=t_.year, doy=t_.julday,
                                          sds_type="D", **tr_.stats)
            full_path = os.path.join(self.tempdir, full_path)
            dirname, filename = os.path.split(full_path)
            if not os.path.isdir(dirname):
                os.makedirs(dirname)
            tr_.write(full_path, format="MSEED")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):  # @UnusedVariable
        os.chdir(self.old_dir)
        shutil.rmtree(self.tempdir)


class TestSDS():
    """
    Test reading data from SDS file structure.
    """
    def test_read_from_sds(self):
        """
        Test reading data across year and day breaks from SDS directory
        structure. Also tests looking for data on the wrong side of a day
        break (data usually get written for some seconds into the previous or
        next file around midnight).
        """
        # test for one specific SEED ID, without wildcards
        seed_id = "AB.XYZ..HHZ"
        net, sta, loc, cha = seed_id.split(".")
        # use three different day breaks in the testing:
        #  - normal day break during one year
        #     (same directory, separate filenames)
        #  - day break at end of year
        #     (separate directories, separate filenames)
        #      - leap-year
        #      - non-leap-year
        for year, doy in ((2015, 123), (2015, 1), (2012, 1)):
            t = UTCDateTime("%d-%03dT00:00:00" % (year, doy))
            with TemporarySDSDirectory(year=year, doy=doy) as temp_sds:
                # normal test reading across the day break
                client = Client(temp_sds.tempdir)
                st = client.get_waveforms(net, sta, loc, cha, t - 20, t + 20)
                assert len(st) == 1
                assert st[0].stats.starttime == t - 20
                assert st[0].stats.endtime == t + 20
                assert len(st[0]) == 5
                # test merge option
                st = client.get_waveforms(net, sta, loc, cha, t - 200, t + 200,
                                          merge=False)
                assert len(st) == 2
                st = client.get_waveforms(net, sta, loc, cha, t - 200, t + 200,
                                          merge=None)
                assert len(st) == 2
                st = client.get_waveforms(net, sta, loc, cha, t - 200, t + 200,
                                          merge=0)
                assert len(st) == 1
                # test reading data from a single day file
                # (data is in the file where it's expected)
                st = client.get_waveforms(net, sta, loc, cha, t - 80, t - 30)
                assert len(st) == 1
                # test reading data from a single day file
                # (data is in the dayfile of the previous day)
                st = client.get_waveforms(net, sta, loc, cha, t + 20, t + 40)
                assert len(st) == 1
                # test that format autodetection with `format=None` works
                client = Client(temp_sds.tempdir, format=None)
                st = client.get_waveforms(net, sta, loc, cha, t - 200, t + 200)
                assert len(st) == 1

    def test_read_from_sds_with_wildcarded_seed_ids(self):
        """
        Test reading data with wildcarded SEED IDs.
        """
        year, doy = 2015, 1
        t = UTCDateTime("%d-%03dT00:00:00" % (year, doy))
        with TemporarySDSDirectory(year=year, doy=doy) as temp_sds:
            # test different wildcard combinations in SEED ID
            client = Client(temp_sds.tempdir)
            for wildcarded_seed_id, num_matching_ids in zip(
                    ("AB.ZZZ3..HH?", "AB.ZZZ3..HH*", "*.*..HHZ",
                     "*.*.*.HHZ", "*.*.*.*"),
                    (3, 3, 4, 8, 48)):
                net, sta, loc, cha = wildcarded_seed_id.split(".")
                st = client.get_waveforms(net, sta, loc, cha, t - 200, t + 200)
                assert len(st) == num_matching_ids
            # test with SDS type wildcards
            for type_wildcard in ("*", "?"):
                net, sta, loc, cha = wildcarded_seed_id.split(".")
                st = client.get_waveforms(net, sta, loc, cha, t - 200, t + 200)
                assert len(st) == num_matching_ids

    def test_sds_report(self, testdata):
        """
        Test command line script for generating SDS report html.

        Inherently that script uses many other routines like `_get_filenames`,
        `get_availability_percentage`, `_get_current_endtime`,
        `get_latency`, `has_data` and `get_all_stations`, so these should be
        sufficiently covered as well.
        """
        # generate some dummy SDS with data roughly 2-3 hours behind current
        # time
        t = UTCDateTime() - 2.5 * 3600
        with TemporarySDSDirectory(year=None, doy=None, time=t) as temp_sds, \
                TemporaryWorkingDirectory():
            # create the report
            output_basename = "sds_report"
            argv = [
                "-r={}".format(temp_sds.tempdir),
                "-o={}".format(os.path.join(os.curdir, output_basename)),
                "-l=", "-l=00", "-l=10", "-c=HHZ", "-c=BHZ", "-i=AB.XYZ..BHE",
                "--check-quality-days=1"]
            sds_report(argv)
            # do the testing
            output_basename_abspath = os.path.abspath(
                os.path.join(os.curdir, output_basename))
            file_html = output_basename_abspath + ".html"
            file_txt = output_basename_abspath + ".txt"
            file_png = output_basename_abspath + ".png"
            # check that output files exist
            for file_ in [file_html, file_txt, file_png]:
                assert os.path.isfile(file_)
            # check content of image file (just check it is a png file)
            # look for png static header
            with open(file_png, 'rb') as fh:
                assert fh.read(8) == b'\x89PNG\r\n\x1a\n'
            # check content of stream info / data quality file
            expected_lines = sorted([
                b"AB,XYZ,,BHE,831[0-9].[0-9]*?,0.007292,2",
                b"AB,XYZ,,HHZ,831[0-9].[0-9]*?,0.007292,2",
                b"AB,XYZ,00,HHZ,831[0-9].[0-9]*?,0.007292,2",
                b"AB,ZZZ3,,HHZ,831[0-9].[0-9]*?,0.007292,2",
                b"AB,ZZZ3,00,HHZ,831[0-9].[0-9]*?,0.007292,2",
                b"CD,XYZ,,HHZ,831[0-9].[0-9]*?,0.007292,2",
                b"CD,XYZ,00,HHZ,831[0-9].[0-9]*?,0.007292,2",
                b"CD,ZZZ3,,HHZ,831[0-9].[0-9]*?,0.007292,2",
                b"CD,ZZZ3,00,HHZ,831[0-9].[0-9]*?,0.007292,2"])
            with open(file_txt, "rb") as fh:
                got_lines = fh.readlines()
            for expected_line, got_line in zip(expected_lines, got_lines):
                assert re.match(expected_line, got_line) is not None
            # check content of html report
            with open(file_html, "rb") as fh:
                got_lines = fh.readlines()
            with open(testdata["sds_report.regex"], "rb") as fh:
                regex_patterns = fh.readlines()
            failed = False  # XXX remove again
            for got, pattern in zip(got_lines, regex_patterns):
                match = re.match(pattern.strip(), got.strip())
                try:
                    assert match is not None
                except AssertionError:
                    failed = True
                    print(pattern.strip())
                    print(got.strip())
            if failed:
                raise Exception

    def test_get_waveforms_bulk(self):
        """
        Test get_waveforms_bulk method.
        """
        year = 2015
        doy = 247
        t = UTCDateTime("%d-%03dT00:00:00" % (year, doy))
        with TemporarySDSDirectory(year=year, doy=doy) as temp_sds:
            chunks = [
                ["AB", "XYZ", "", "HHZ", t, t + 20],
                ["AB", "XYZ", "", "HHN", t + 20, t + 40],
                ["AB", "XYZ", "", "HHE", t + 40, t + 60],
                ["CD", "ZZZ3", "00", "BHZ", t + 60, t + 80],
                ["CD", "ZZZ3", "00", "BHN", t + 80, t + 100],
                ["CD", "ZZZ3", "00", "BHE", t + 120, t + 140]
            ]
            client = Client(temp_sds.tempdir)
            st = client.get_waveforms_bulk(chunks)
            for _i in range(6):
                if _i <= 2:
                    assert st[_i].stats.network == "AB"
                    assert st[_i].stats.station == "XYZ"
                    assert st[_i].stats.location == ""
                elif _i >= 2:
                    assert st[_i].stats.network == "CD"
                    assert st[_i].stats.station == "ZZZ3"
                    assert st[_i].stats.location == "00"
            assert st[0].stats.channel == "HHZ"
            assert st[1].stats.channel == "HHN"
            assert st[2].stats.channel == "HHE"
            assert st[3].stats.channel == "BHZ"
            assert st[4].stats.channel == "BHN"
            assert st[5].stats.channel == "BHE"

    def test_get_all_stations_and_nslc(self):
        """
        Test `get_all_stations` and `get_all_nslc` methods
        """
        # generate dummy SDS
        t = UTCDateTime()
        with TemporarySDSDirectory(year=None, doy=None, time=t) as temp_sds:
            client = Client(temp_sds.tempdir)
            expected_netsta = sorted([
                (net, sta)
                for net in temp_sds.networks
                for sta in temp_sds.stations])
            got_netsta = client.get_all_stations()
            assert expected_netsta == got_netsta
            expected_nslc = sorted([
                (net, sta, loc, cha)
                for net in temp_sds.networks
                for sta in temp_sds.stations
                for loc in temp_sds.locations
                for cha in temp_sds.channels])
            got_nslc = client.get_all_nslc()
            assert expected_nslc == got_nslc
            got_nslc = client.get_all_nslc(datetime=t)
            assert expected_nslc == got_nslc
            # other dates that have no data should return empty list
            got_nslc = client.get_all_nslc(datetime=t + 2 * 24 * 3600)
            assert [] == got_nslc
            got_nslc = client.get_all_nslc(datetime=t - 2 * 24 * 3600)
            assert [] == got_nslc



import io
from urllib.request import urlopen

import numpy as np
import matplotlib.pyplot as plt
from scipy.cluster import hierarchy
from scipy.spatial import distance

from obspy.imaging.cm import obspy_sequential


url = "https://examples.obspy.org/dissimilarities.npz"
with io.BytesIO(urlopen(url).read()) as fh, np.load(fh) as data:
    dissimilarity = data['dissimilarity']

plt.subplot(121)
plt.imshow(1 - dissimilarity, interpolation='nearest', cmap=obspy_sequential)

dissimilarity = distance.squareform(dissimilarity)
threshold = 0.3
linkage = hierarchy.linkage(dissimilarity, method="single")
clusters = hierarchy.fcluster(linkage, threshold, criterion="distance")

# A little nicer set of colors.
cmap = plt.get_cmap('Paired', lut=6)
colors = ['#%02x%02x%02x' % tuple(int(col * 255) for col in cmap(i)[:3])
          for i in range(6)]
try:
    hierarchy.set_link_color_palette(colors[1:])
except AttributeError:
    # Old version of SciPy
    pass

plt.subplot(122)
try:
    hierarchy.dendrogram(linkage, color_threshold=0.3,
                         above_threshold_color=cmap(0))
except TypeError:
    # Old version of SciPy
    hierarchy.dendrogram(linkage, color_threshold=0.3)
plt.xlabel("Event number")
plt.ylabel("Dissimilarity")
plt.show()



# -*- coding: utf-8 -*-
"""
Library name handling for ObsPy.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import ctypes
import importlib.machinery
from pathlib import Path
import re


def cleanse_pymodule_filename(filename):
    """
    Replace all characters not allowed in Python module names in filename with
    "_".

    See bug report:
     - https://stackoverflow.com/q/21853678
     - See #755

    See also:
     - https://stackoverflow.com/q/7552311
     - https://docs.python.org/3/reference/lexical_analysis.html#identifiers

    >>> print(cleanse_pymodule_filename("0blup-bli.554_3!32"))
    _blup_bli_554_3_32
    """
    filename = re.sub(r'^[^a-zA-Z_]', "_", filename)
    filename = re.sub(r'[^a-zA-Z0-9_]', "_", filename)
    return filename


def _load_cdll(name):
    """
    Helper function to load a shared library built during ObsPy installation
    with ctypes.

    :type name: str
    :param name: Name of the library to load (e.g. 'mseed').
    :rtype: :class:`ctypes.CDLL`
    """
    errors = []
    libdir = Path(__file__).parent.parent.parent / 'lib'
    for ext in importlib.machinery.EXTENSION_SUFFIXES:
        libpath = (libdir / (name + ext)).resolve()
        try:
            cdll = ctypes.CDLL(str(libpath))
        except Exception as e:
            errors.append(f'    {str(e)}')
        else:
            return cdll
    # If we got here, then none of the attempted extensions worked.
    raise ImportError('\n  '.join([
        f'Could not load shared library "{name}"',
        *errors,
        'Current directory: %s' % Path().resolve(),
        'Directory listing of lib directory:',
        *(f'    {str(d)}' for d in sorted(libpath.parent.iterdir())),
    ]))


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



# -*- coding: utf-8 -*-
import os
import re
import tempfile
import uuid
from collections import namedtuple
from pathlib import Path
from unittest import mock

import pytest
import sqlalchemy as sa
from sqlalchemy.orm import sessionmaker

from obspy import read, UTCDateTime
from obspy.clients.filesystem.tsindex import Client, Indexer, \
    TSIndexDatabaseHandler


# this one is used on the API frontpage, so guess leave it in for now parallel
# to the proper fixture used in testing
def get_test_data_filepath():
    package_dir = os.path.abspath(os.path.dirname(__file__))
    # TODO It perhaps shouldn't be the case, but the string here has to end
    # in a forward slash. We should fix this in the future.
    filepath = os.path.join(package_dir, 'data', 'tsindex_data/')
    return filepath


@pytest.fixture(scope="module")
def filepath(request):
    """
    Dictionary with full paths to additional tsindex test files in
    subdirectory of "/data" for convenience.
    """
    module_path = Path(request.module.__file__)
    datapath = module_path.parent / 'data' / 'tsindex_data'
    return datapath


@pytest.fixture(scope="function")
def client(filepath):
    db_path = filepath / 'timeseries.sqlite'

    client = Client(str(db_path), datapath_replace=("^", str(filepath) + '/'))
    return client


class TestClient():

    def test_bad_sqlitdb_filepath(self):
        """
        Checks that an error is raised when an invalid path is provided to
        a SQLite database
        """
        with pytest.raises(OSError, match="^Database path.*does not exist.$"):
            Client("/some/bad/path/timeseries.sqlite")

    def test_get_waveforms(self, filepath, client):
        path = (filepath / 'IU' / '2018' / '001' /
                'IU.ANMO.10.BHZ.2018.001_first_minute.mseed')
        expected_stream = read(
            path, starttime=UTCDateTime(2018, 1, 1, 0, 0, 0),
            endtime=UTCDateTime(2018, 1, 1, 0, 0, 5))
        returned_stream = client.get_waveforms(
            "IU", "ANMO", "10", "BHZ",
            starttime=UTCDateTime(2018, 1, 1, 0, 0, 0),
            endtime=UTCDateTime(2018, 1, 1, 0, 0, 5))
        returned_stream.sort()
        expected_stream.sort()

        for t1, t2 in zip(returned_stream, expected_stream):
            assert list(t1.data) == list(t2.data)

        # wildcard request spanning multiple files
        expected_stream1 = \
            read(
                filepath / 'CU' / '2018' / '001' /
                'CU.TGUH.00.BHZ.2018.001_first_minute.mseed',
                starttime=UTCDateTime(2018, 1, 1, 0, 0, 0),
                endtime=UTCDateTime(2018, 1, 1, 0, 0, 3, 1))
        expected_stream2 = \
            read(
                filepath / 'IU' / '2018' / '001' /
                'IU.ANMO.10.BHZ.2018.001_first_minute.mseed',
                starttime=UTCDateTime(2018, 1, 1, 0, 0, 0),
                endtime=UTCDateTime(2018, 1, 1, 0, 0, 3, 1))
        expected_stream3 = \
            read(
                filepath / 'IU' / '2018' / '001' /
                'IU.COLA.10.BHZ.2018.001_first_minute.mseed',
                starttime=UTCDateTime(2018, 1, 1, 0, 0, 0),
                endtime=UTCDateTime(2018, 1, 1, 0, 0, 3, 1))
        expected_stream = \
            expected_stream1 + expected_stream2 + expected_stream3
        returned_stream = client.get_waveforms(
                                  "*", "*", "10,00", "BHZ",
                                  starttime=UTCDateTime(2018, 1, 1, 0, 0, 0),
                                  endtime=UTCDateTime(2018, 1, 1, 0, 0, 3, 1))
        returned_stream.sort()
        expected_stream.sort()

        assert len(returned_stream) == len(expected_stream)
        for t1, t2 in zip(returned_stream, expected_stream):
            assert list(t1.data) == list(t2.data)

        # request resulting in no data
        returned_stream = client.get_waveforms(
                                  "XX", "XXX", "XX", "XXX",
                                  starttime=UTCDateTime(2018, 1, 1, 0, 0, 0),
                                  endtime=UTCDateTime(2018, 1, 1, 0, 0, 3, 1))
        assert returned_stream.traces == []

    def test_get_waveforms_bulk(self, filepath, client):
        expected_stream1 = \
            read(
                filepath /
                'CU/2018/001/CU.TGUH.00.BHZ.2018.001_first_minute.mseed',
                starttime=UTCDateTime(2018, 1, 1, 0, 0, 1),
                endtime=UTCDateTime(2018, 1, 1, 0, 0, 7))
        expected_stream2 = \
            read(
                filepath /
                'IU/2018/001/IU.ANMO.10.BHZ.2018.001_first_minute.mseed',
                starttime=UTCDateTime(2018, 1, 1, 0, 0, 0),
                endtime=UTCDateTime(2018, 1, 1, 0, 0, 5))
        expected_stream = expected_stream1 + expected_stream2
        expected_stream.sort()

        # assert equal
        bulk_request = [
                        ("CU", "TGUH", "00", "BHZ",
                         UTCDateTime(2018, 1, 1, 0, 0, 1),
                         UTCDateTime(2018, 1, 1, 0, 0, 7)),
                        ("IU", "ANMO", "10", "BHZ",
                         UTCDateTime(2018, 1, 1, 0, 0, 0),
                         UTCDateTime(2018, 1, 1, 0, 0, 5)),
                        ]
        returned_stream = client.get_waveforms_bulk(bulk_request)
        returned_stream.sort()

        assert len(returned_stream) == len(expected_stream)
        for t1, t2 in zip(returned_stream, expected_stream):
            assert list(t1.data) == list(t2.data)

        # assert not equal with a non-equivalent request
        bulk_request = [
                        ("IU", "ANMO", "10", "BHZ",
                         UTCDateTime(2018, 1, 1, 0, 0, 0),
                         UTCDateTime(2018, 1, 1, 0, 0, 5)),
                        ]
        returned_stream = client.get_waveforms_bulk(bulk_request)
        returned_stream.sort()
        assert len(returned_stream) != len(expected_stream)
        for t1, t2 in zip(returned_stream, expected_stream):
            with pytest.raises(AssertionError):
                assert list(t1.data) == list(t2.data)

        # request resulting in no data
        returned_stream = client.get_waveforms(
                                  "XX", "XXX", "XX", "XXX",
                                  starttime=UTCDateTime(2018, 1, 1, 0, 0, 0),
                                  endtime=UTCDateTime(2018, 1, 1, 0, 0, 3, 1))
        assert returned_stream.traces == []

    def test_get_nslc(self, client):
        # test using actual sqlite3 test database
        expected_nslc = [(u'CU', u'TGUH', u'00', u'BHZ')]

        actual_nslc = client.get_nslc("I*,CU",
                                      "ANMO,COL?,T*",
                                      "00,10",
                                      "BHZ",
                                      "2018-01-01T00:00:00.000000",
                                      "2018-01-01T00:00:00.019499")
        assert actual_nslc == expected_nslc
        actual_nslc = client.get_nslc("CU",
                                      "ANMO,COL?,T*")
        assert actual_nslc == expected_nslc

        # test using mocked client._get_summary_rows method for more diversity
        NamedRow = namedtuple('NamedRow',
                              ['network', 'station', 'location', 'channel',
                               'earliest', 'latest'])
        mocked_summary_rows = [NamedRow("AK", "ANM", "", "VM2",
                                        "2018-08-10T21:52:50.000000",
                                        "2018-08-10T22:12:39.999991"),
                               NamedRow("AK", "ANM", "", "VM3",
                                        "2018-08-10T21:52:50.000000",
                                        "2018-08-10T22:12:39.999991"),
                               NamedRow("AK", "ANM", "", "VM4",
                                        "2018-08-10T21:52:50.000000",
                                        "2018-08-10T22:12:39.999991"),
                               NamedRow("XX", "ANM", "", "VM5",
                                        "2018-08-10T21:52:50.000000",
                                        "2018-08-10T22:12:39.999991"),
                               NamedRow("N4", "H43A", "", "VM2",
                                        "2018-08-10T21:09:39.000000",
                                        "2018-08-10T22:09:28.890415"),
                               NamedRow("N4", "H43A", "", "VM3",
                                        "2018-08-10T21:09:39.000000",
                                        "2018-08-10T22:09:28.890415")]
        client._get_summary_rows = \
            mock.MagicMock(return_value=mocked_summary_rows)

        expected_nslc = [("AK", "ANM", "", "VM2"),
                         ("AK", "ANM", "", "VM3"),
                         ("AK", "ANM", "", "VM4"),
                         ("N4", "H43A", "", "VM2"),
                         ("N4", "H43A", "", "VM3"),
                         ("XX", "ANM", "", "VM5")]

        assert client.get_nslc(
            "AK,N4,XX", "ANM,H43A", "", "VM2,VM3,VM4,VM5",
            "2018-08-10T21:09:39.000000",
            "2018-08-10T22:09:28.890415") == expected_nslc

    def test_get_availability_extent(self, client):
        # test using actual sqlite test database
        expected_nslc = [(u'IU', u'ANMO', u'10', u'BHZ',
                          UTCDateTime(2018, 1, 1, 0, 0, 0, 19500),
                          UTCDateTime(2018, 1, 1, 0, 0, 59, 994536)),
                         (u'IU', u'COLA', u'10', u'BHZ',
                          UTCDateTime(2018, 1, 1, 0, 0, 0, 19500),
                          UTCDateTime(2018, 1, 1, 0, 0, 59, 994538))]

        actual_avail_extents = client.get_availability_extent(
                                                "I*",
                                                "ANMO,COL?,T*",
                                                "00,10",
                                                "BHZ",
                                                "2018-01-01T00:00:00.000000",
                                                "2018-12-31T00:00:00.000000")
        assert actual_avail_extents == expected_nslc

        actual_avail_extents = client.get_availability_extent("I*")
        assert actual_avail_extents == expected_nslc

        # test using mocked client._get_summary_rows method for more diversity
        NamedRow = namedtuple('NamedRow',
                              ['network', 'station', 'location', 'channel',
                               'earliest', 'latest'])
        mocked_summary_rows = [NamedRow("AK", "ANM", "", "VM2",
                                        "2018-08-10T21:52:50.000000",
                                        "2018-08-10T22:12:39.999991"),
                               NamedRow("AK", "ANM", "", "VM3",
                                        "2018-08-10T21:52:50.000000",
                                        "2018-08-10T22:12:39.999991"),
                               NamedRow("AK", "ANM", "", "VM5",
                                        "2018-08-10T21:52:50.000000",
                                        "2018-08-10T22:15:39.999991"),
                               NamedRow("N4", "H43A", "", "VM2",
                                        "2018-08-10T21:09:39.000000",
                                        "2018-08-10T22:09:28.890415"),
                               NamedRow("N4", "H43A", "", "VM3",
                                        "2018-08-10T21:09:39.000000",
                                        "2018-08-10T22:09:28.890415"),
                               NamedRow("XX", "ANM", "", "VM4",
                                        "2018-08-10T21:52:50.000000",
                                        "2018-08-10T22:12:39.999991")]
        client._get_summary_rows = \
            mock.MagicMock(return_value=mocked_summary_rows)

        expected_avail_extents = \
            [("AK", "ANM", "", "VM2",
              UTCDateTime("2018-08-10T21:52:50.000000"),
              UTCDateTime("2018-08-10T22:12:39.999991")),
             ("AK", "ANM", "", "VM3",
              UTCDateTime("2018-08-10T21:52:50.000000"),
              UTCDateTime("2018-08-10T22:12:39.999991")),
             ("AK", "ANM", "", "VM5",
              UTCDateTime("2018-08-10T21:52:50.000000"),
              UTCDateTime("2018-08-10T22:15:39.999991")),
             ("N4", "H43A", "", "VM2",
              UTCDateTime("2018-08-10T21:09:39.000000"),
              UTCDateTime("2018-08-10T22:09:28.890415")),
             ("N4", "H43A", "", "VM3",
              UTCDateTime("2018-08-10T21:09:39.000000"),
              UTCDateTime("2018-08-10T22:09:28.890415")),
             ("XX", "ANM", "", "VM4",
              UTCDateTime("2018-08-10T21:52:50.000000"),
              UTCDateTime("2018-08-10T22:12:39.999991"))]

        assert client.get_availability_extent(
            "AK,N4", "ANM,H43A", "", "VM2,VM3,VM4,VM5",
            "2018-08-10T21:09:39.000000",
            "2018-08-10T22:09:28.890415") == expected_avail_extents

    def test__are_timespans_adjacent(self, client):
        sample_rate = 40
        # sample_period = 1/40 = 0.025 sec = 25000 ms
        # and tolerance = 0.5 so an adjacent sample is +/-0.0125 sec = 12500 ms

        # 1 sample period later
        ts1 = \
            client._create_timespan(UTCDateTime(2018, 8, 10, 22, 0, 0, 0),
                                    UTCDateTime(2018, 8, 10, 22, 0, 0, 25000))
        ts2 = \
            client._create_timespan(UTCDateTime(2018, 8, 10, 22, 0, 0, 50000),
                                    UTCDateTime(2018, 8, 10, 22, 0, 0, 75000))
        assert client._are_timespans_adjacent(ts1, ts2, sample_rate)

        # 1ms after nearest tolerance boundary (next sample - 12500ms)
        ts1 = \
            client._create_timespan(UTCDateTime(2018, 8, 10, 22, 0, 0, 0),
                                    UTCDateTime(2018, 8, 10, 22, 0, 0, 25000))
        ts2 = \
            client._create_timespan(UTCDateTime(2018, 8, 10, 22, 0, 0, 37501),
                                    UTCDateTime(2018, 8, 10, 22, 0, 0, 75000))
        assert client._are_timespans_adjacent(ts1, ts2, sample_rate)

        # exactly on nearest tolerance boundary (next sample - 12500ms)
        ts1 = \
            client._create_timespan(UTCDateTime(2018, 8, 10, 22, 0, 0, 0),
                                    UTCDateTime(2018, 8, 10, 22, 0, 0, 25000))
        ts2 = \
            client._create_timespan(UTCDateTime(2018, 8, 10, 22, 0, 0, 37500),
                                    UTCDateTime(2018, 8, 10, 22, 0, 0, 75000))
        assert not client._are_timespans_adjacent(ts1, ts2, sample_rate)

        # 1ms before nearest tolerance boundary (next sample - 12500ms)
        ts1 = \
            client._create_timespan(UTCDateTime(2018, 8, 10, 22, 0, 0, 0),
                                    UTCDateTime(2018, 8, 10, 22, 0, 0, 25000))
        ts2 = \
            client._create_timespan(UTCDateTime(2018, 8, 10, 22, 0, 0, 37499),
                                    UTCDateTime(2018, 8, 10, 22, 0, 0, 75000))
        assert not client._are_timespans_adjacent(ts1, ts2, sample_rate)

        # 1ms after farthest tolerance boundary (next sample + 12500ms)
        ts1 = \
            client._create_timespan(UTCDateTime(2018, 8, 10, 22, 0, 0, 0),
                                    UTCDateTime(2018, 8, 10, 22, 0, 0, 25000))
        ts2 = \
            client._create_timespan(UTCDateTime(2018, 8, 10, 22, 0, 0, 62501),
                                    UTCDateTime(2018, 8, 10, 22, 0, 0, 100000))
        assert not client._are_timespans_adjacent(ts1, ts2, sample_rate)

        # on farthest tolerance boundary (next sample + 12500ms)
        ts1 = \
            client._create_timespan(UTCDateTime(2018, 8, 10, 22, 0, 0, 0),
                                    UTCDateTime(2018, 8, 10, 22, 0, 0, 25000))
        ts2 = \
            client._create_timespan(UTCDateTime(2018, 8, 10, 22, 0, 0, 62500),
                                    UTCDateTime(2018, 8, 10, 22, 0, 0, 100000))
        assert not client._are_timespans_adjacent(ts1, ts2, sample_rate)

        # 1ms before farthest tolerance boundary (next sample + 12500ms)
        ts1 = \
            client._create_timespan(UTCDateTime(2018, 8, 10, 22, 0, 0, 0),
                                    UTCDateTime(2018, 8, 10, 22, 0, 0, 25000))
        ts2 = \
            client._create_timespan(UTCDateTime(2018, 8, 10, 22, 0, 0, 62499),
                                    UTCDateTime(2018, 8, 10, 22, 0, 0, 100000))
        assert client._are_timespans_adjacent(ts1, ts2, sample_rate)

    def test_get_availability(self, client):
        # test using actual sqlite test database
        expected_avail = [(u'IU', u'ANMO', u'10', u'BHZ',
                           UTCDateTime(2018, 1, 1, 0, 0, 0, 19500),
                           UTCDateTime(2018, 1, 1, 0, 0, 59, 994536)),
                          (u'IU', u'COLA', u'10', u'BHZ',
                           UTCDateTime(2018, 1, 1, 0, 0, 0, 19500),
                           UTCDateTime(2018, 1, 1, 0, 0, 59, 994538))]

        actual_avail = client.get_availability(
                                                "IU",
                                                "ANMO,COLA",
                                                "10",
                                                "BHZ",
                                                "2018-01-01",
                                                "2018-12-31")
        assert actual_avail == expected_avail

        actual_avail = client.get_availability("IU",
                                               "ANMO,COLA")
        assert actual_avail == expected_avail

        # test using mocked client._get_summary_rows method for more diversity
        NamedRow = namedtuple('NamedRow',
                              ['network', 'station', 'location', 'channel',
                               'samplerate', 'starttime', 'endtime',
                               'timespans'])
        # Each row contains one gap. The first rows latest time is adjacent
        # to the second rows earliest time. The third row overlaps with the
        # second row.
        mocked_tsindex_rows = \
            [
               # 2018-08-10T22:00:54 to 2018-08-10T22:15:53
               # 2018-08-10T22:05:54 to 2018-08-10T22:20:53 (MERGE)
               NamedRow(
                   network=u'AK', station=u'BAGL',
                   location=u'', channel=u'LCC',
                   starttime=u'2018-08-10T22:00:54.000000',
                   endtime=u'2018-08-10T22:20:53.000000',
                   samplerate=1.0,
                   timespans=u'[1533938454.000000:1533939353.000000],'
                             u'[1533938754.000000:1533939653.000000]'),
               # 2018-08-10T22:20:53.999000 to 2018-08-12T22:20:53 (JOIN)
               # 2018-08-12T23:20:53 to 2018-09-01T23:20:53
               NamedRow(
                   network=u'AK', station=u'BAGL',
                   location=u'', channel=u'LCC',
                   starttime=u'2018-08-10T22:20:53.999000',
                   endtime=u'2018-09-01T23:20:53.000000',
                   samplerate=1.0,
                   timespans=u'[1533939653.999000:1534112453.000000],'
                             u'[1534116053.000000:1535844053.000000]'),
               # (MERGE IF INCL SAMPLE RATE IS TRUE)
               # 2018-08-27T00:00:00 to 2018-09-11T00:00:00
               NamedRow(network=u'AK', station=u'BAGL',
                        location=u'', channel=u'LCC',
                        starttime=u'2018-08-27T00:00:00.000000',
                        endtime=u'2018-09-11T00:00:00.000000',
                        samplerate=10.0,
                        timespans=u'[1535328000.0:1536624000.0]')
            ]
        client._get_tsindex_rows = \
            mock.MagicMock(return_value=mocked_tsindex_rows)

        expected_unmerged_avail = [("AK", "BAGL", "", "LCC",
                                    UTCDateTime(2018, 8, 10, 22, 0, 54),
                                    UTCDateTime(2018, 8, 10, 22, 15, 53)),
                                   ("AK", "BAGL", "", "LCC",
                                    UTCDateTime(2018, 8, 10, 22, 5, 54),
                                    UTCDateTime(2018, 8, 12, 22, 20, 53)),
                                   ("AK", "BAGL", "", "LCC",
                                    UTCDateTime(2018, 8, 12, 23, 20, 53),
                                    UTCDateTime(2018, 9, 1, 23, 20, 53)),
                                   ("AK", "BAGL", "", "LCC",
                                    UTCDateTime(2018, 8, 27, 0, 0),
                                    UTCDateTime(2018, 9, 11, 0, 0, 0))]

        # test default options
        assert client.get_availability(
            "AK", "BAGL", "", "LCC",
            UTCDateTime(2018, 8, 10, 22, 0, 54),
            UTCDateTime(2018, 8, 10, 22, 9, 28, 890415)) == \
            expected_unmerged_avail

        # test merge overlap false
        assert client.get_availability(
            "AK", "BAGL", "--", "LCC",
            UTCDateTime(2018, 8, 10, 22, 0, 54),
            UTCDateTime(2018, 8, 10, 22, 9, 28, 890415),
            merge_overlap=False) == expected_unmerged_avail

        # test merge overlap true
        expected_merged_avail = [("AK", "BAGL", "", "LCC",
                                  UTCDateTime(2018, 8, 10, 22, 0, 54),
                                  UTCDateTime(2018, 8, 12, 22, 20, 53)),
                                 ("AK", "BAGL", "", "LCC",
                                  UTCDateTime(2018, 8, 12, 23, 20, 53),
                                  UTCDateTime(2018, 9, 11, 0, 0, 0))]

        assert client.get_availability(
            "AK", "BAGL", "--", "LCC",
            UTCDateTime(2018, 8, 10, 22, 0, 54),
            UTCDateTime(2018, 8, 10, 22, 9, 28, 890415),
            merge_overlap=True) == expected_merged_avail

        # test include_sample_rate true
        expected_incl_sr_avail = \
            [("AK", "BAGL", "", "LCC",
              UTCDateTime(2018, 8, 10, 22, 0, 54),
              UTCDateTime(2018, 8, 12, 22, 20, 53), 1.0),
             ("AK", "BAGL", "", "LCC",
              UTCDateTime(2018, 8, 12, 23, 20, 53),
              UTCDateTime(2018, 9, 1, 23, 20, 53), 1.0),
             ("AK", "BAGL", "", "LCC",
              UTCDateTime(2018, 8, 27, 0, 0),
              UTCDateTime(2018, 9, 11, 0, 0, 0), 10.0)]

        assert client.get_availability(
            "AK", "BAGL", "--", "LCC",
            UTCDateTime(2018, 8, 10, 22, 0, 54),
            UTCDateTime(2018, 8, 10, 22, 9, 28, 890415),
            merge_overlap=True,
            include_sample_rate=True) == expected_incl_sr_avail

    def test_get_availability_percentage(self, client):
        mock_availability_output = [("AK", "BAGL", "", "LCC",
                                     UTCDateTime(2018, 8, 10, 22, 0, 54),
                                     UTCDateTime(2018, 8, 12, 22, 20, 53)),
                                    ("AK", "BAGL", "", "LCC",
                                     UTCDateTime(2018, 8, 12, 23, 20, 53),
                                     UTCDateTime(2018, 9, 11, 0, 0, 0))]
        client.get_availability = mock.MagicMock(
            return_value=mock_availability_output)

        avail_percentage = client.get_availability_percentage(
                                    "AK",
                                    "BAGL",
                                    "--",
                                    "LCC",
                                    UTCDateTime(2018, 8, 10, 22, 0, 54),
                                    UTCDateTime(2018, 9, 11, 0, 0, 0))
        expected_avail_percentage = (0.998659490472, 1)
        assert round(
            abs(avail_percentage[0]-expected_avail_percentage[0]), 7) == 0
        assert avail_percentage[1] == expected_avail_percentage[1]
        assert isinstance(avail_percentage, tuple)

        mock_availability_output = [("AK", "BAGL", "", "LCC",
                                     UTCDateTime(2018, 1, 1),
                                     UTCDateTime(2018, 1, 2)),
                                    ("AK", "BAGL", "", "LCC",
                                     UTCDateTime(2018, 1, 3),
                                     UTCDateTime(2018, 1, 4)),
                                    ("AK", "BAGL", "", "LCC",
                                     UTCDateTime(2018, 1, 5),
                                     UTCDateTime(2018, 1, 6))]
        client.get_availability = mock.MagicMock(
            return_value=mock_availability_output)

        avail_percentage = client.get_availability_percentage(
                                                "AK",
                                                "BAGL",
                                                "--",
                                                "LCC",
                                                UTCDateTime(2018, 1, 1),
                                                UTCDateTime(2018, 1, 6))
        expected_avail_percentage = (0.6, 2)
        assert round(
            abs(avail_percentage[0]-expected_avail_percentage[0]), 7) == 0
        assert avail_percentage[1] == expected_avail_percentage[1]
        assert isinstance(avail_percentage, tuple)

        # Test for over extending time span
        avail_percentage = client.get_availability_percentage(
                                                "AK",
                                                "BAGL",
                                                "--",
                                                "LCC",
                                                UTCDateTime(2017, 12, 31),
                                                UTCDateTime(2018, 1, 7))
        expected_avail_percentage = (0.4285714, 4)
        assert round(
            abs(avail_percentage[0]-expected_avail_percentage[0]), 7) == 0
        assert avail_percentage[1] == expected_avail_percentage[1]
        assert isinstance(avail_percentage, tuple)

    def test_has_data(self, client):
        assert client.has_data()
        assert client.has_data(starttime=UTCDateTime(2017, 12, 31),
                               endtime=UTCDateTime(2018, 1, 7))
        assert not client.has_data(starttime=UTCDateTime(1970, 12, 31),
                                   endtime=UTCDateTime(2013, 1, 7))


def purge(dir, pattern):
    """
    Delete a file matching a pattern from the OS.
    """
    for f in os.listdir(dir):
        if re.search(pattern, f):
            os.remove(os.path.join(dir, f))


class TestIndexer():

    def test_bad_rootpath(self, filepath):
        """
        Checks that an error is raised when an invalid path is provided to
        a data
        """
        database = os.path.join(filepath,
                                'timeseries.sqlite')

        # test that a bad root path raises an error
        with pytest.raises(OSError, match="^Root path.*does not exists.$"):
            Indexer("/some/bad/path", database=database,
                    filename_pattern="*.mseed", parallel=2)

    def test_bad_sqlitdb_filepath(self, filepath):
        """
        Checks that an error is raised when an invalid path is provided to
        a SQLite database
        """
        with pytest.raises(OSError, match="^Database path.*does not exist.$"):
            Indexer(filepath, database='/some/bad/path/',
                    filename_pattern="*.mseed")

    def test_bad_database(self, filepath):
        """
        Checks that an error is raised when an invalid database is provided.
        The database must be a TSIndexDatabaseHandler or a str, otherwise
        a ValueError is raised.
        """
        match = "^Database must be a string or TSIndexDatabaseHandler object.$"
        with pytest.raises(ValueError, match=match):
            Indexer(filepath, database=None, filename_pattern="*.mseed")

    def test_build_file_list(self, filepath):
        database = os.path.join(filepath, 'timeseries.sqlite')
        indexer = Indexer(filepath,
                          database=database,
                          filename_pattern="*.mseed")

        # test for relative paths
        file_list = indexer.build_file_list(relative_paths=True,
                                            reindex=True)
        file_list.sort()
        assert len(file_list) == 3
        assert os.path.normpath(
            'CU/2018/001/CU.TGUH.00.BHZ.2018.001_first_minute.mseed') == \
            file_list[0]
        assert os.path.normpath(
            'IU/2018/001/IU.ANMO.10.BHZ.2018.001_first_minute.mseed') == \
            file_list[1]
        assert os.path.normpath(
            'IU/2018/001/IU.COLA.10.BHZ.2018.001_first_minute.mseed') == \
            file_list[2]

        # case where the root path is outside of the absolute
        # data path, to assert that already indexed files are still skipped
        indexer = Indexer(tempfile.mkdtemp(),
                          database=TSIndexDatabaseHandler(database=database),
                          filename_pattern="*.mseed")
        with pytest.raises(OSError, match="^No files matching filename.*$"):
            indexer.build_file_list(reindex=True)

        # test for absolute paths
        # this time pass a TSIndexDatabaseHandler instance as the database
        indexer = Indexer(filepath,
                          database=TSIndexDatabaseHandler(database=database),
                          filename_pattern="*.mseed")
        file_list = indexer.build_file_list(reindex=True)
        file_list.sort()
        assert len(file_list) == 3
        assert os.path.normpath(
            'CU/2018/001/CU.TGUH.00.BHZ.2018.001_first_minute.mseed') != \
            file_list[0]
        assert os.path.normpath(
            'CU/2018/001/CU.TGUH.00.BHZ.2018.001_first_minute.mseed') in \
            file_list[0]
        assert os.path.normpath(
            'IU/2018/001/IU.ANMO.10.BHZ.2018.001_first_minute.mseed') != \
            file_list[1]
        assert os.path.normpath(
            'IU/2018/001/IU.ANMO.10.BHZ.2018.001_first_minute.mseed') in \
            file_list[1]
        assert os.path.normpath(
            'IU/2018/001/IU.COLA.10.BHZ.2018.001_first_minute.mseed') != \
            file_list[2]
        assert os.path.normpath(
            'IU/2018/001/IU.COLA.10.BHZ.2018.001_first_minute.mseed') in \
            file_list[2]
        # test that already indexed files (relative and absolute) get skipped.
        with pytest.raises(
                OSError, match="^No unindexed files matching filename.*$"):
            indexer.build_file_list(reindex=False, relative_paths=False)
        with pytest.raises(
                OSError, match="^No unindexed files matching filename.*$"):
            indexer.build_file_list(reindex=False, relative_paths=True)
        # for this test mock an unindexed file ('data.mseed') to ensure that
        # it gets added when reindex is True
        mocked_files = [
                'CU/2018/001/'
                'CU.TGUH.00.BHZ.2018.001_first_minute.mseed',
                'IU/2018/001/'
                'IU.ANMO.10.BHZ.2018.001_first_minute.mseed',
                'IU/2018/001/'
                'IU.COLA.10.BHZ.2018.001_first_minute.mseed',
                'data.mseed'
            ]
        for i in range(len(mocked_files)):
            mocked_files[i] = os.path.normpath(mocked_files[i])
        indexer._get_rootpath_files = mock.MagicMock(return_value=mocked_files)
        assert indexer.build_file_list(
            reindex=False, relative_paths=False) == ['data.mseed']

    def test_run_bad_index_cmd(self, filepath):
        """
        Checks that an OSError is raised when there is an error running a
        index_cmd. (such as no command found.)
        """
        indexer = Indexer(filepath, filename_pattern="*.mseed",
                          index_cmd="some_bad_command")

        match = "^Required program.* is not installed.*$"
        with pytest.raises(OSError, match=match):
            indexer.run()

    def test_run(self, filepath):
        my_uuid = uuid.uuid4().hex
        fname = 'test_timeseries_{}'.format(my_uuid)
        database = '{}{}.sqlite'.format(filepath, fname)
        try:
            indexer = Indexer(filepath,
                              database=database,
                              filename_pattern="*.mseed",
                              parallel=2)
            if indexer._is_index_cmd_installed():
                indexer.run(relative_paths=True)
                keys = ['network', 'station', 'location', 'channel',
                        'quality', 'starttime', 'endtime', 'samplerate',
                        'filename', 'byteoffset', 'bytes', 'hash',
                        'timeindex', 'timespans', 'timerates', 'format']
                NamedRow = namedtuple('NamedRow',
                                      keys)

                expected_tsindex_data = \
                    [
                     NamedRow(
                        "CU", "TGUH", "00", "BHZ", None,
                        "2018-01-01T00:00:00",
                        "2018-01-01T00:01:00", 40.0,
                        "CU/2018/001/"
                        "CU.TGUH.00.BHZ.2018.001_first_minute.mseed",
                        0, 4096, "aaaac5315f84cdd174fd8360002a1e3a",
                        "1514764800.000000=>0,latest=>1",
                        "[1514764800.000000:1514764860.000000]", None, None),
                     NamedRow(
                        "IU", "ANMO", "10", "BHZ", None,
                        "2018-01-01T00:00:00.019500",
                        "2018-01-01T00:00:59.994536", 40.0,
                        "IU/2018/001/"
                        "IU.ANMO.10.BHZ.2018.001_first_minute.mseed",
                        0, 2560, "36a771ca1dc648c505873c164d8b26f2",
                        "1514764800.019500=>0,latest=>1",
                        "[1514764800.019500:1514764859.994536]", None, None),
                     NamedRow(
                        "IU", "COLA", "10", "BHZ", None,
                        "2018-01-01T00:00:00.019500",
                        "2018-01-01T00:00:59.994538", 40.0,
                        "IU/2018/001/"
                        "IU.COLA.10.BHZ.2018.001_first_minute.mseed",
                        0, 5120, "4ccbb97573ca00ef8c2c4f9c01d27ddf",
                        "1514764800.019500=>0,latest=>1",
                        "[1514764800.019500:1514764859.994538]", None, None)]
                db_handler = TSIndexDatabaseHandler(database=database)
                tsindex_data = db_handler._fetch_index_rows([("I*,C*", "*",
                                                              "0?,1?", "*",
                                                              "2018-01-01",
                                                              "2018-02-01")])

                for i in range(0, len(expected_tsindex_data)):
                    for j in range(0, len(keys)):
                        assert getattr(
                            expected_tsindex_data[i], keys[j]) == \
                            getattr(tsindex_data[i], keys[j])
                assert len(tsindex_data) == len(expected_tsindex_data)
        finally:
            purge(filepath, '^{}.*$'.format(fname))


class TestTSIndexDatabaseHandler():

    def test_bad_sqlitdb_filepath(self, filepath):
        """
        Checks that an error is raised when an invalid path is provided to
        a SQLite database
        """
        with pytest.raises(OSError, match="^Database path.*does not exist.$"):
            Indexer(filepath, database='/some/bad/path/',
                    filename_pattern="*.mseed", parallel=2)

    def test__fetch_summary_rows(self, filepath):
        # test with actual sqlite3 database that is missing a summary table
        # a temporary summary table gets created at runtime
        db_path = os.path.join(filepath, 'timeseries.sqlite')
        request_handler = TSIndexDatabaseHandler(db_path)

        keys = ['network', 'station', 'location', 'channel',
                'earliest', 'latest']
        NamedRow = namedtuple('NamedRow',
                              keys)

        expected_ts_summary_data = \
            [NamedRow(
                "CU", "TGUH", "00", "BHZ",
                "2018-01-01T00:00:00.000000",
                "2018-01-01T00:01:00.000000"),
             NamedRow(
                "IU", "ANMO", "10", "BHZ",
                "2018-01-01T00:00:00.019500",
                "2018-01-01T00:00:59.994536")]

        ts_summary_data = request_handler._fetch_summary_rows(
                                              [("I*,CU",
                                                "ANMO,T*",
                                                "00,10",
                                                "BHZ",
                                                "2018-01-01T00:00:00.000000",
                                                "2018-12-31T00:00:00.000000")])

        for i in range(0, len(expected_ts_summary_data)):
            for j in range(0, len(keys)):
                assert getattr(expected_ts_summary_data[i], keys[j]) == \
                                 getattr(ts_summary_data[i], keys[j])

        # test for case where query returns no results
        ts_summary_data = request_handler._fetch_summary_rows(
            [("XX", "ANMO,T*", "00,10", "BHZ",
              "2018-01-01T00:00:00.000000",
              "2018-12-31T00:00:00.000000")])
        assert ts_summary_data == []

    def test_get_tsindex_summary_cte(self, filepath):
        # test with actual sqlite3 database that is missing a summary table
        # a tsindex summary CTE gets created using the tsindex at runtime
        db_path = os.path.join(filepath, 'timeseries.sqlite')
        # supply an existing session
        engine = sa.create_engine("sqlite:///{}".format(db_path))
        session = sessionmaker(bind=engine)
        request_handler = TSIndexDatabaseHandler(session=session)

        ts_summary_cte = request_handler.get_tsindex_summary_cte()

        expected_ts_summary_data = \
            [("CU", "TGUH", "00", "BHZ",
              "2018-01-01T00:00:00.000000",
              "2018-01-01T00:01:00.000000"),
             ("IU", "ANMO", "10", "BHZ",
              "2018-01-01T00:00:00.019500",
              "2018-01-01T00:00:59.994536"),
             ("IU", "COLA", "10", "BHZ",
              "2018-01-01T00:00:00.019500",
              "2018-01-01T00:00:59.994538")]
        query_results = (session().query(ts_summary_cte))
        for idx, r in enumerate(query_results):
            result = r[:6]  # ignore updt date
            assert result == expected_ts_summary_data[idx]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
The paz test suite.
"""
import io

from obspy.io.gse2 import paz


class TestPAZ():
    """
    Test cases for reading GSE PAZ files.
    """
    def test_read_with_spaces(self):
        """
        Reading PAZ files where PAZ are separated by spaces.
        """
        f = io.StringIO("""CAL1 RJOB   LE-3D    Z  M24    PAZ 010824 0001
        2
        -4.39823 4.48709
        -4.39823 -4.48709
        3
        0.0 0.0
        0.0 0.0
        0.0 0.0
        0.4""")
        p, z, k = paz.read_paz(f)
        assert round(abs(-4.39823-p[0].real), 7) == 0
        assert round(abs(4.48709-p[0].imag), 7) == 0
        assert round(abs(-4.39823-p[1].real), 7) == 0
        assert round(abs(-4.48709-p[1].imag), 7) == 0
        assert [0j, 0j, 0j] == z
        assert round(abs(0.4-k), 7) == 0
        f.close()

    def test_read_without_spaces(self):
        """
        Reading PAZ files where PAZ are not separated by spaces.

        Tests uses unrealistic PAZ information.
        """
        f = io.StringIO("""CAL1 RJOB   LE-3D    Z  M24    PAZ 010824 0001
2
-4.3982340.48709
-4.39823-4.48709
3
1.2 4.0
-1.09823-3.08709
-1.0982330.08709
0.5""")
        p, z, k = paz.read_paz(f)
        assert round(abs(-4.39823-p[0].real), 7) == 0
        assert round(abs(40.48709-p[0].imag), 7) == 0
        assert round(abs(-4.39823-p[1].real), 7) == 0
        assert round(abs(-4.48709-p[1].imag), 7) == 0
        assert round(abs(1.2-z[0].real), 7) == 0
        assert round(abs(4.0-z[0].imag), 7) == 0
        assert round(abs(-1.09823-z[1].real), 7) == 0
        assert round(abs(-3.08709-z[1].imag), 7) == 0
        assert round(abs(-1.09823-z[2].real), 7) == 0
        assert round(abs(30.08709-z[2].imag), 7) == 0
        assert round(abs(0.5-k), 7) == 0
        f.close()



# -*- coding: utf-8 -*-
import io
import os
import pickle
import warnings

import numpy as np
import pytest

from obspy import UTCDateTime, read_events
from obspy.core.event import (Catalog, Comment, CreationInfo, Event,
                              FocalMechanism, Magnitude, Origin, Pick,
                              ResourceIdentifier, WaveformStreamID)
from obspy.core.event.source import farfield
from obspy.core.util import CARTOPY_VERSION
from obspy.core.util.base import _get_entry_points, CatchAndAssertWarnings
from obspy.core.util.misc import MatplotlibBackend
from obspy.core.event.base import QuantityError


if CARTOPY_VERSION and CARTOPY_VERSION >= [0, 12, 0]:
    HAS_CARTOPY = True
else:
    HAS_CARTOPY = False


class TestEvent:
    """
    Test suite for obspy.core.event.event.Event
    """

    def test_str(self):
        """
        Testing the __str__ method of the Event object.
        """
        event = read_events()[1]
        s = event.short_str()
        expected = ("2012-04-04T14:18:37.000000Z | +39.342,  +41.044" +
                    " | 4.3  ML | manual")
        assert s == expected

    def test_str_empty_origin(self):
        """
        Ensure an event with an empty origin returns a str without raising a
        TypeError (#2119).
        """
        event = Event(origins=[Origin()])
        out = event.short_str()
        assert isinstance(out, str)
        assert out == 'None | None, None'

    def test_eq(self):
        """
        Testing the __eq__ method of the Event object.
        """
        # events are equal if the have the same public_id
        # Catch warnings about the same different objects with the same
        # resource id so they do not clutter the test output.
        with warnings.catch_warnings() as _:  # NOQA
            warnings.simplefilter("ignore")
            ev1 = Event(resource_id='id1')
            ev2 = Event(resource_id='id1')
            ev3 = Event(resource_id='id2')
        assert ev1 == ev2
        assert ev2 == ev1
        assert ev1 != ev3
        assert ev3 != ev1
        # comparing with other objects fails
        assert ev1 != 1
        assert ev2 != "id1"

    def test_clear_method_resets_objects(self):
        """
        Tests that the clear() method properly resets all objects. Test for
        #449.
        """
        # Test with basic event object.
        e = Event(force_resource_id=False)
        e.comments.append(Comment(text="test"))
        e.event_type = "explosion"
        assert len(e.comments) == 1
        assert e.event_type == "explosion"
        e.clear()
        assert e == Event(force_resource_id=False)
        assert len(e.comments) == 0
        assert e.event_type is None
        # Test with pick object. Does not really fit in the event test case but
        # it tests the same thing...
        p = Pick()
        p.comments.append(Comment(text="test"))
        p.phase_hint = "p"
        assert len(p.comments) == 1
        assert p.phase_hint == "p"
        # Add some more random attributes. These should disappear upon
        # cleaning.
        with CatchAndAssertWarnings() as w:
            p.test_1 = "a"
            p.test_2 = "b"
            # two warnings should have been issued by setting non-default keys
            assert len(w) == 2
        assert p.test_1 == "a"
        assert p.test_2 == "b"
        p.clear()
        assert len(p.comments) == 0
        assert p.phase_hint is None
        assert not hasattr(p, "test_1")
        assert not hasattr(p, "test_2")

    @pytest.mark.skipif(not (CARTOPY_VERSION and CARTOPY_VERSION >=
                             [0, 12, 0]),
                        reason='cartopy not installed')
    def test_plot_farfield_without_quiver_with_maps(self, image_path):
        """
        Tests to plot P/S wave farfield radiation pattern, also with beachball
        and some map plots.
        """
        ev = read_events("/path/to/CMTSOLUTION", format="CMTSOLUTION")[0]
        ev.plot(kind=[['global'], ['ortho', 'beachball'],
                      ['p_sphere', 's_sphere']], outfile=image_path)

    def test_farfield_2xn_input(self):
        """
        Tests to compute P/S wave farfield radiation pattern using (theta,phi)
        pairs as input
        """
        # Peru 2001/6/23 20:34:23:
        #  RTP system: [2.245, -0.547, -1.698, 1.339, -3.728, 1.444]
        #  NED system: [-0.547, -1.698, 2.245, -1.444, 1.339, 3.728]
        mt = [-0.547, -1.698, 2.245, -1.444, 1.339, 3.728]
        theta = np.arange(0, 360, 60)
        phi = np.zeros(len(theta))
        rays = np.array([theta, phi]) * np.pi / 180.0
        result = farfield(mt, rays, 'P')
        ref = np.array([[0., 1.13501984, -0.873480164, 2.749332e-16,
                        -1.13501984, 0.873480164], [0, 0, -0, 0, -0, 0],
                        [2.245, 0.655304008, 0.504304008, -2.245,
                         -0.655304008, -0.504304008]])
        np.testing.assert_allclose(result, ref, rtol=1e-5, atol=1e-8)


class TestOrigin:
    """
    Test suite for obspy.core.event.Origin
    """
    def test_creation_info(self):
        # 1 - empty Origin class will set creation_info to None
        orig = Origin()
        assert orig.creation_info is None
        # 2 - preset via dict or existing CreationInfo object
        orig = Origin(creation_info={})
        assert isinstance(orig.creation_info, CreationInfo)
        orig = Origin(creation_info=CreationInfo(author='test2'))
        assert isinstance(orig.creation_info, CreationInfo)
        assert orig.creation_info.author == 'test2'
        # 3 - check set values
        orig = Origin(creation_info={'author': 'test'})
        assert orig.creation_info == orig['creation_info']
        assert orig.creation_info.author == 'test'
        assert orig['creation_info']['author'] == 'test'
        orig.creation_info.agency_id = "muh"
        assert orig.creation_info == orig['creation_info']
        assert orig.creation_info.agency_id == 'muh'
        assert orig['creation_info']['agency_id'] == 'muh'

    def test_multiple_origins(self):
        """
        Parameters of multiple origins should not interfere with each other.
        """
        origin = Origin()
        origin.resource_id = 'smi:ch.ethz.sed/origin/37465'
        origin.time = UTCDateTime(0)
        origin.latitude = 12
        origin.latitude_errors.confidence_level = 95
        origin.longitude = 42
        origin.depth_type = 'from location'
        assert origin.resource_id == \
            ResourceIdentifier(id='smi:ch.ethz.sed/origin/37465')
        assert origin.latitude == 12
        assert origin.latitude_errors.confidence_level == 95
        assert origin.latitude_errors.uncertainty is None
        assert origin.longitude == 42
        origin2 = Origin(force_resource_id=False)
        origin2.latitude = 13.4
        assert origin2.depth_type is None
        assert origin2.resource_id is None
        assert origin2.latitude == 13.4
        assert origin2.latitude_errors.confidence_level is None
        assert origin2.longitude is None


class TestCatalog:
    """
    Test suite for obspy.core.event.Catalog
    """
    def test_read_invalid_filename(self):
        """
        Tests that we get a sane error message when calling read_events()
        with a filename that doesn't exist
        """
        doesnt_exist = 'dsfhjkfs'
        for i in range(10):
            if os.path.exists(doesnt_exist):
                doesnt_exist += doesnt_exist
                continue
            break
        else:
            self.fail('unable to get invalid file path')

        exception_msg = "No such file or directory"

        formats = _get_entry_points(
            'obspy.plugin.catalog', 'readFormat').keys()
        # try read_inventory() with invalid filename for all registered read
        # plugins and also for filetype autodiscovery
        formats = [None] + list(formats)
        for format in formats:
            with pytest.raises(FileNotFoundError, match=exception_msg):
                read_events(doesnt_exist, format=format)

    def test_creation_info(self):
        cat = Catalog()
        cat.creation_info = CreationInfo(author='test2')
        assert isinstance(cat.creation_info, CreationInfo)
        assert cat.creation_info.author == 'test2'

    def test_read_events_without_parameters(self):
        """
        Calling read_events w/o any parameter will create an example catalog.
        """
        catalog = read_events()
        assert len(catalog) == 3

    def test_str(self):
        """
        Testing the __str__ method of the Catalog object.
        """
        catalog = read_events()
        assert catalog.__str__().startswith("3 Event(s) in Catalog:")
        assert catalog.__str__().endswith(
            "37.736 | 3.0  ML | manual")

    def test_read_events(self, testdata):
        """
        Tests the read_events() function using entry points.
        """
        # iris
        catalog = read_events(testdata['iris_events.xml'])
        assert len(catalog) == 2
        assert catalog[0]._format == 'QUAKEML'
        assert catalog[1]._format == 'QUAKEML'
        # neries
        catalog = read_events(testdata['neries_events.xml'])
        assert len(catalog) == 3
        assert catalog[0]._format == 'QUAKEML'
        assert catalog[1]._format == 'QUAKEML'
        assert catalog[2]._format == 'QUAKEML'

    def test_read_events_with_wildcard(self, datapath, testdata):
        """
        Tests the read_events() function with a filename wild card.
        """
        # without wildcard..
        expected = read_events(testdata['iris_events.xml'])
        expected += read_events(testdata['neries_events.xml'])
        # with wildcard
        got = read_events(datapath / "*_events.xml")
        assert expected == got

    def test_append(self):
        """
        Tests the append method of the Catalog object.
        """
        # 1 - create catalog and add a few events
        catalog = Catalog()
        event1 = Event()
        event2 = Event()
        assert len(catalog) == 0
        catalog.append(event1)
        assert len(catalog) == 1
        assert catalog.events == [event1]
        catalog.append(event2)
        assert len(catalog) == 2
        assert catalog.events == [event1, event2]
        # 2 - adding objects other as Event should fails
        with pytest.raises(TypeError):
            catalog.append(str)
        with pytest.raises(TypeError):
            catalog.append(Catalog)
        with pytest.raises(TypeError):
            catalog.append([event1])

    def test_extend(self):
        """
        Tests the extend method of the Catalog object.
        """
        # 1 - create catalog and extend it with list of events
        catalog = Catalog()
        event1 = Event()
        event2 = Event()
        assert len(catalog) == 0
        catalog.extend([event1, event2])
        assert len(catalog) == 2
        assert catalog.events == [event1, event2]
        # 2 - extend it with other catalog
        event3 = Event()
        event4 = Event()
        catalog2 = Catalog([event3, event4])
        assert len(catalog) == 2
        catalog.extend(catalog2)
        assert len(catalog) == 4
        assert catalog.events == [event1, event2, event3, event4]
        # adding objects other as Catalog or list should fails
        with pytest.raises(TypeError):
            catalog.extend(str)
        with pytest.raises(TypeError):
            catalog.extend(event1)
        with pytest.raises(TypeError):
            catalog.extend((event1, event2))

    def test_iadd(self):
        """
        Tests the __iadd__ method of the Catalog object.
        """
        # 1 - create catalog and add it with another catalog
        event1 = Event()
        event2 = Event()
        event3 = Event()
        catalog = Catalog([event1])
        catalog2 = Catalog([event2, event3])
        assert len(catalog) == 1
        catalog += catalog2
        assert len(catalog) == 3
        assert catalog.events == [event1, event2, event3]
        # 3 - extend it with another Event
        event4 = Event()
        assert len(catalog) == 3
        catalog += event4
        assert len(catalog) == 4
        assert catalog.events == [event1, event2, event3, event4]
        # adding objects other as Catalog or Event should fails
        with pytest.raises(TypeError):
            catalog.__iadd__(str)
        with pytest.raises(TypeError):
            catalog.__iadd__((event1, event2))
        with pytest.raises(TypeError):
            catalog.__iadd__([event1, event2])

    def test_count_and_len(self):
        """
        Tests the count and __len__ methods of the Catalog object.
        """
        # empty catalog without events
        catalog = Catalog()
        assert len(catalog) == 0
        assert catalog.count() == 0
        # catalog with events
        catalog = read_events()
        assert len(catalog) == 3
        assert catalog.count() == 3

    def test_get_item(self):
        """
        Tests the __getitem__ method of the Catalog object.
        """
        catalog = read_events()
        assert catalog[0] == catalog.events[0]
        assert catalog[-1] == catalog.events[-1]
        assert catalog[2] == catalog.events[2]
        # out of index should fail
        with pytest.raises(IndexError):
            catalog.__getitem__(3)
        with pytest.raises(IndexError):
            catalog.__getitem__(-99)

    def test_slicing(self):
        """
        Tests the __getslice__ method of the Catalog object.
        """
        catalog = read_events()
        assert catalog[0:] == catalog[0:]
        assert catalog[:2] == catalog[:2]
        assert catalog[:] == catalog[:]
        assert len(catalog) == 3
        new_catalog = catalog[1:3]
        assert isinstance(new_catalog, Catalog)
        assert len(new_catalog) == 2

    def test_slicing_with_step(self):
        """
        Tests the __getslice__ method of the Catalog object with step.
        """
        ev1 = Event()
        ev2 = Event()
        ev3 = Event()
        ev4 = Event()
        ev5 = Event()
        catalog = Catalog([ev1, ev2, ev3, ev4, ev5])
        assert catalog[0:6].events == [ev1, ev2, ev3, ev4, ev5]
        assert catalog[0:6:1].events == [ev1, ev2, ev3, ev4, ev5]
        assert catalog[0:6:2].events == [ev1, ev3, ev5]
        assert catalog[1:6:2].events == [ev2, ev4]
        assert catalog[1:6:6].events == [ev2]

    def test_copy(self):
        """
        Testing the copy method of the Catalog object.
        """
        cat = read_events()
        cat2 = cat.copy()
        assert cat == cat2
        assert cat2 == cat
        assert not (cat is cat2)
        assert not (cat2 is cat)
        assert cat.events[0] == cat2.events[0]
        assert not (cat.events[0] is cat2.events[0])

    def test_filter(self):
        """
        Testing the filter method of the Catalog object.
        """
        def getattrs(event, attr):
            if attr == 'magnitude':
                obj = event.magnitudes[0]
                attr = 'mag'
            else:
                obj = event.origins[0]
            for a in attr.split('.'):
                obj = getattr(obj, a)
            return obj
        cat = read_events()
        assert all(event.magnitudes[0].mag < 4.
                   for event in cat.filter('magnitude < 4.'))
        attrs = ('magnitude', 'latitude', 'longitude', 'depth', 'time',
                 'quality.standard_error', 'quality.azimuthal_gap',
                 'quality.used_station_count', 'quality.used_phase_count')
        values = (4., 40., 50., 10., UTCDateTime('2012-04-04 14:20:00'),
                  1., 50, 40, 20)
        for attr, value in zip(attrs, values):
            attr_filter = attr.split('.')[-1]
            cat_smaller = cat.filter('%s < %s' % (attr_filter, value))
            cat_bigger = cat.filter('%s >= %s' % (attr_filter, value))
            assert all(True if a is None else a < value
                       for event in cat_smaller
                       for a in [getattrs(event, attr)])
            assert all(False if a is None else a >= value
                       for event in cat_bigger
                       for a in [getattrs(event, attr)])
            assert all(event in cat for event in (cat_smaller + cat_bigger))
            cat_smaller_inverse = cat.filter(
                '%s < %s' % (attr_filter, value), inverse=True)
            assert all(event in cat_bigger for event in cat_smaller_inverse)
            cat_bigger_inverse = cat.filter(
                '%s >= %s' % (attr_filter, value), inverse=True)
            assert all(event in cat_smaller for event in cat_bigger_inverse)

    def test_catalog_resource_id(self, testdata):
        """
        See #662
        """
        cat = read_events(testdata['neries_events.xml'])
        assert str(cat.resource_id) == r"smi://eu.emsc/unid"

    def test_can_pickle(self):
        """
        Ensure a catalog can be pickled and unpickled and that the results are
        equal.
        """
        cat = read_events()
        cat_bytes = pickle.dumps(cat)
        cat2 = pickle.loads(cat_bytes)
        assert cat == cat2

    def test_issue_2173(self):
        """
        Ensure events with empty origins are equal after round-trip to disk.
        See #2173.
        """
        # create event and save to disk
        origin = Origin(time=UTCDateTime('2016-01-01'))
        event1 = Event(origins=[origin])
        bio = io.BytesIO()
        event1.write(bio, 'quakeml')
        # rewind bytes stream
        bio.seek(0)
        # read from disk
        event2 = read_events(bio)[0]
        # saved and loaded event should be equal
        assert event1 == event2

    def test_read_path(self, testdata):
        """
        Ensure read_events works with pathlib.Path objects.
        """
        assert read_events(str(testdata['iris_events.xml'])) == \
            read_events(testdata['iris_events.xml'])


@pytest.mark.skipif(not HAS_CARTOPY,
                    reason='Cartopy not installed or too old')
class CatalogCartopyTestCase:
    """
    Test suite for obspy.core.event.Catalog.plot using Cartopy
    """
    def test_catalog_plot_global(self, image_path):
        """
        Tests the catalog preview plot, default parameters, using Cartopy.
        """
        cat = read_events()
        cat.plot(method='cartopy', outfile=image_path)

    def test_catalog_plot_ortho(self, image_path):
        """
        Tests the catalog preview plot, ortho projection, some non-default
        parameters, using Cartopy.
        """
        cat = read_events()
        cat.plot(method='cartopy', outfile=image_path, projection='ortho',
                 resolution='c', water_fill_color='#98b7e2', label=None,
                 color='date')

    def test_catalog_plot_ortho_longitude_wrap(self, image_path):
        """
        Tests the catalog preview plot, ortho projection, some non-default
        parameters, using Cartopy, with longitudes that need the mean to be
        computed in a circular fashion.
        """
        cat = read_events('/path/to/events_longitude_wrap.zmap', format='ZMAP')
        cat.plot(method='cartopy', outfile=image_path, projection='ortho',
                 resolution='c', label=None, title='', colorbar=False,
                 water_fill_color='b')

    def test_catalog_plot_local(self, image_path):
        """
        Tests the catalog preview plot, local projection, some more non-default
        parameters, using Cartopy.
        """
        cat = read_events()
        cat.plot(method='cartopy', outfile=image_path, projection='local',
                 resolution='50m', continent_fill_color='0.3',
                 color='date', colormap='gist_heat')

    def test_plot_catalog_before_1900(self):
        """
        Tests plotting events with origin times before 1900
        """
        cat = read_events()
        cat[1].origins[0].time = UTCDateTime(813, 2, 4, 14, 13)

        # just checking this runs without error is fine, no need to check
        # content
        with MatplotlibBackend("AGG", sloppy=True):
            cat.plot(outfile=io.BytesIO())
            # also test with just a single event
            cat.events = [cat[1]]
            cat.plot(outfile=io.BytesIO())


class TestWaveformStreamID:
    """
    Test suite for obspy.core.event.WaveformStreamID.
    """
    def test_initialization(self):
        """
        Test the different initialization methods.
        """
        # Default init.
        waveform_id = WaveformStreamID()
        assert waveform_id.network_code is None
        assert waveform_id.station_code is None
        assert waveform_id.location_code is None
        assert waveform_id.channel_code is None
        # With seed string.
        waveform_id = WaveformStreamID(seed_string="BW.FUR.01.EHZ")
        assert waveform_id.network_code == "BW"
        assert waveform_id.station_code == "FUR"
        assert waveform_id.location_code == "01"
        assert waveform_id.channel_code == "EHZ"
        # As soon as any other argument is set, the seed_string will not be
        # used and the default values will be used for any unset arguments.
        waveform_id = WaveformStreamID(location_code="02",
                                       seed_string="BW.FUR.01.EHZ")
        assert waveform_id.network_code is None
        assert waveform_id.station_code is None
        assert waveform_id.location_code == "02"
        assert waveform_id.channel_code is None

    def test_initialization_with_invalid_seed_string(self):
        """
        Test initialization with an invalid seed string. Should raise a
        warning.
        """
        # An invalid SEED string will issue a warning and fill the object with
        # the default values.
        with warnings.catch_warnings(record=True):
            warnings.simplefilter('error', UserWarning)
            with pytest.raises(UserWarning):
                WaveformStreamID(seed_string="Invalid SEED string")
            # Now ignore the warnings and test the default values.
            warnings.simplefilter('ignore', UserWarning)
            waveform_id = WaveformStreamID(seed_string="Invalid Seed String")
            assert waveform_id.network_code is None
            assert waveform_id.station_code is None
            assert waveform_id.location_code is None
            assert waveform_id.channel_code is None

    def test_id_property(self):
        """
        Enure the `id` property of WaveformStreamID returns the same as
        `get_seed_string`"
        """
        waveform_id = WaveformStreamID(seed_string="BW.FUR.01.EHZ")
        assert waveform_id.id == waveform_id.get_seed_string()


class TestBase:
    """
    Test suite for obspy.core.event.base.
    """
    def test_quantity_error_warn_on_non_default_key(self):
        """
        """
        err = QuantityError()
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            err.uncertainty = 0.01
            err.lower_uncertainty = 0.1
            err.upper_uncertainty = 0.02
            err.confidence_level = 80
            assert len(w) == 0
            # setting a typoed or custom field should warn!
            err.confidence_levle = 80
            assert len(w) == 1

    def test_quantity_error_equality(self):
        """
        Comparisons between empty quantity errors and None should return True.
        Non-empty quantity errors should return False.
        """
        err1 = QuantityError()
        assert err1 == None  # NOQA needs to be ==
        err2 = QuantityError(uncertainty=10)
        assert err2 is not None
        assert err2 != err1
        err3 = QuantityError(uncertainty=10)
        assert err3 == err2

    def test_event_type_objects_warn_on_non_default_key(self):
        """
        """
        for cls in (Event, Origin, Pick, Magnitude, FocalMechanism):
            obj = cls()
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always")
                # setting a typoed or custom field should warn!
                obj.some_custom_non_default_crazy_key = "my_text_here"
                assert len(w) == 1

    def test_setting_nans_or_inf_fails(self):
        """
        Tests that settings NaNs or infs as floating point values fails.
        """
        o = Origin()

        with pytest.raises(ValueError, match='is not a finite'):
            o.latitude = float('nan')

        with pytest.raises(ValueError, match='is not a finite'):
            o.latitude = float('inf')

        with pytest.raises(ValueError, match='is not a finite'):
            o.latitude = float('-inf')

    def test_issue3105(self):
        evs = read_events()
        evs[0].magnitudes[0].mag = 0
        assert len(evs) == 3
        assert len(evs.filter('magnitude < 3.5')) == 2



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Tests for differentiation and integration functions.
"""

import numpy as np

from obspy.signal.differentiate_and_integrate import (
    integrate_cumtrapz, integrate_spline)


class TestIntegrate():
    """
    Test cases for the integration methods.
    """
    def test_cumtrapz_integration(self):
        """
        Test the basic and obvious cases. We are using external methods which
        are tested extensively elsewhere.
        """
        np.testing.assert_allclose(
            integrate_cumtrapz(np.ones(10), dx=1.0),
            np.arange(10))

        np.testing.assert_allclose(
            integrate_cumtrapz(np.ones(10), dx=2.0),
            np.arange(10) * 2.0)

        np.testing.assert_allclose(
            integrate_cumtrapz(np.ones(10), dx=0.5),
            np.arange(10) * 0.5)

        np.testing.assert_allclose(
            integrate_cumtrapz(np.zeros(10), dx=0.5),
            np.zeros(10))

    def test_spline_integration(self):
        """
        Test the basic and obvious cases. We are using external methods which
        are tested extensively elsewhere.
        """
        for k in (1, 2, 3):
            np.testing.assert_allclose(
                integrate_spline(np.ones(10), dx=1.0, k=k),
                np.arange(10))

            np.testing.assert_allclose(
                integrate_spline(np.ones(10), dx=2.0, k=k),
                np.arange(10) * 2.0)

            np.testing.assert_allclose(
                integrate_spline(np.ones(10), dx=0.5, k=k),
                np.arange(10) * 0.5)

            np.testing.assert_allclose(
                integrate_spline(np.zeros(10), dx=0.5, k=k),
                np.zeros(10))



import obspy
from obspy.signal.invsim import corn_freq_2_paz


paz_sts2 = {
    'poles': [-0.037004 + 0.037016j, -0.037004 - 0.037016j, -251.33 + 0j,
              - 131.04 - 467.29j, -131.04 + 467.29j],
    'zeros': [0j, 0j],
    'gain': 60077000.0,
    'sensitivity': 2516778400.0}
paz_1hz = corn_freq_2_paz(1.0, damp=0.707)  # 1Hz instrument
paz_1hz['sensitivity'] = 1.0

st = obspy.read()
# make a copy to keep our original data
st_orig = st.copy()

# Simulate instrument given poles, zeros and gain of
# the original and desired instrument
st.simulate(paz_remove=paz_sts2, paz_simulate=paz_1hz)

# plot original and simulated data
st_orig.plot()
st.plot()



# -*- coding: utf-8 -*-
import os
import warnings

import numpy as np

from obspy import Trace, UTCDateTime, read
from obspy.core.util import NamedTemporaryFile
from obspy.io.sh.core import (STANDARD_ASC_HEADERS, _is_asc, _is_q, _read_asc,
                              _read_q, _write_asc, _write_q)


class TestCore():
    """
    """
    def test_read_101_traces(self, testdata):
        """
        Testing reading Q file with more than 100 traces.
        """
        testfile = testdata['101.QHD']
        # read
        stream = _read_q(testfile)
        stream.verify()
        assert len(stream) == 101

    def test_is_asc_file(self, testdata):
        """
        Testing ASC file format.
        """
        testfile = testdata['TEST_090101_0101.ASC']
        assert _is_asc(testfile)
        testfile = testdata['QFILE-TEST-SUN.QHD']
        assert not _is_asc(testfile)

    def test_is_q_file(self, testdata):
        """
        Testing Q header file format.
        """
        testfile = testdata['QFILE-TEST-SUN.QHD']
        assert _is_q(testfile)
        testfile = testdata['QFILE-TEST-SUN.QBN']
        assert not _is_q(testfile)
        testfile = testdata['TEST_090101_0101.ASC']
        assert not _is_q(testfile)

    def test_read_single_channel_asc_file(self, testdata):
        """
        Read ASC file test via obspy.io.sh.core._read_asc.
        """
        testfile = testdata['TEST_090101_0101.ASC']
        # read
        stream = _read_asc(testfile)
        stream.verify()
        assert stream[0].stats.delta == 5.000000e-02
        assert stream[0].stats.npts == 1000
        assert stream[0].stats.starttime == UTCDateTime(2009, 1, 1, 1, 1, 1)
        assert stream[0].stats.channel == 'BHZ'
        assert stream[0].stats.station == 'TEST'
        assert stream[0].stats.calib == 1.0e-00
        # check last 4 samples
        data = [2.176000e+01, 2.195485e+01, 2.213356e+01, 2.229618e+01]
        np.testing.assert_array_almost_equal(stream[0].data[-4:], data)

    def _compare_stream(self, stream):
        """
        Helper function to verify stream from file 'data/QFILE-TEST*'.
        """
        # channel 1
        assert stream[0].stats.delta == 5.000000e-02
        assert stream[0].stats.npts == 801
        assert stream[0].stats.sh.COMMENT == 'TEST TRACE IN QFILE #1'
        assert stream[0].stats.starttime == UTCDateTime(2009, 10, 1, 12, 46, 1)
        assert stream[0].stats.channel == 'BHN'
        assert stream[0].stats.station == 'TEST'
        assert stream[0].stats.calib == 1.500000e+00
        # check last 4 samples
        data = [-4.070354e+01, -4.033876e+01, -3.995153e+01, -3.954230e+01]
        np.testing.assert_array_almost_equal(stream[0].data[-4:], data, 5)
        # channel 2
        assert stream[1].stats.delta == 5.000000e-02
        assert stream[1].stats.npts == 801
        assert stream[1].stats.sh.COMMENT == 'TEST TRACE IN QFILE #2'
        assert stream[1].stats.starttime == UTCDateTime(2009, 10, 1, 12, 46, 1)
        assert stream[1].stats.channel == 'BHE'
        assert stream[1].stats.station == 'TEST'
        assert stream[1].stats.calib == 1.500000e+00
        # check first 4 samples
        data = [-3.995153e+01, -4.033876e+01, -4.070354e+01, -4.104543e+01]
        np.testing.assert_array_almost_equal(stream[1].data[0:4], data, 5)
        # channel 3
        assert stream[2].stats.delta == 1.000000e-02
        assert stream[2].stats.npts == 4001
        assert stream[2].stats.sh.COMMENT == '******'
        assert stream[2].stats.starttime == \
            UTCDateTime(2010, 1, 1, 1, 1, 5, 999000)
        assert stream[2].stats.channel == 'HHZ'
        assert stream[2].stats.station == 'WET'
        assert stream[2].stats.calib == 1.059300e+00
        # check first 4 samples
        data = [4.449060e+02, 4.279572e+02, 4.120677e+02, 4.237200e+02]
        np.testing.assert_array_almost_equal(stream[2].data[0:4], data, 4)

    def test_read_and_write_multi_channel_asc_file(self, testdata):
        """
        Read and write ASC file via obspy.io.sh.core._read_asc.
        """
        origfile = testdata['QFILE-TEST-ASC.ASC']
        # read original
        stream1 = _read_asc(origfile)
        stream1.verify()
        self._compare_stream(stream1)
        # write
        with NamedTemporaryFile() as tf:
            tempfile = tf.name
            _write_asc(stream1, tempfile, STANDARD_ASC_HEADERS + ['COMMENT'])
            # read both files and compare the content
            with open(origfile, 'rt') as f:
                text1 = f.readlines()
            with open(tempfile, 'rt') as f:
                text2 = f.readlines()
            assert text1 == text2
            # read again
            stream2 = _read_asc(tempfile)
            stream2.verify()
            self._compare_stream(stream2)

    def test_read_and_write_multi_channel_asc_file_via_obspy(self, testdata):
        """
        Read and write ASC file test via obspy.core.
        """
        origfile = testdata['QFILE-TEST-ASC.ASC']
        # read original
        stream1 = read(origfile, format="SH_ASC")
        stream1.verify()
        self._compare_stream(stream1)
        # write
        with NamedTemporaryFile() as tf:
            tempfile = tf.name
            hd = STANDARD_ASC_HEADERS + ['COMMENT']
            stream1.write(tempfile, format="SH_ASC", included_headers=hd)
            # read again w/ auto detection
            stream2 = read(tempfile)
            stream2.verify()
            self._compare_stream(stream2)

    def test_read_and_write_multi_channel_q_file(self, testdata):
        """
        Read and write Q file via obspy.io.sh.core._read_q.
        """
        # 1 - little endian (PC)
        origfile = testdata['QFILE-TEST.QHD']
        # read original
        stream1 = _read_q(origfile)
        stream1.verify()
        self._compare_stream(stream1)
        # write
        with NamedTemporaryFile(suffix='.QHD') as tf:
            tempfile = tf.name
            _write_q(stream1, tempfile, append=False)
            # read again
            stream2 = _read_q(tempfile)
            stream2.verify()
            self._compare_stream(stream2)
            # remove binary file too (dynamically created)
            os.remove(os.path.splitext(tempfile)[0] + '.QBN')
        # 2 - big endian (SUN)
        origfile = testdata['QFILE-TEST-SUN.QHD']
        # read original
        stream1 = _read_q(origfile, byteorder=">")
        stream1.verify()
        self._compare_stream(stream1)
        # write
        with NamedTemporaryFile(suffix='.QHD') as tf:
            tempfile = tf.name
            _write_q(stream1, tempfile, byteorder=">", append=False)
            # read again
            stream2 = _read_q(tempfile, byteorder=">")
            stream2.verify()
            self._compare_stream(stream2)
            # remove binary file too (dynamically created)
            os.remove(os.path.splitext(tempfile)[0] + '.QBN')

    def test_read_and_write_multi_channel_q_file_via_obspy(self, testdata):
        """
        Read and write Q file test via obspy.core.
        """
        # 1 - little endian (PC)
        origfile = testdata['QFILE-TEST.QHD']
        # read original
        stream1 = read(origfile, format="Q")
        stream1.verify()
        self._compare_stream(stream1)
        # write
        with NamedTemporaryFile(suffix='.QHD') as tf:
            tempfile = tf.name
            stream1.write(tempfile, format="Q", append=False)
            # read again w/ auto detection
            stream2 = read(tempfile)
            stream2.verify()
            self._compare_stream(stream2)
            # remove binary file too (dynamically created)
            os.remove(os.path.splitext(tempfile)[0] + '.QBN')
        # 2 - big endian (SUN)
        origfile = testdata['QFILE-TEST-SUN.QHD']
        # read original
        stream1 = read(origfile, format="Q", byteorder=">")
        stream1.verify()
        self._compare_stream(stream1)
        # write
        with NamedTemporaryFile(suffix='.QHD') as tf:
            tempfile = tf.name
            stream1.write(tempfile, format="Q", byteorder=">", append=False)
            # read again w/ auto detection
            stream2 = read(tempfile, byteorder=">")
            stream2.verify()
            self._compare_stream(stream2)
            # remove binary file too (dynamically created)
            os.remove(os.path.splitext(tempfile)[0] + '.QBN')

    def test_skip_asc_lines(self, testdata):
        testfile = testdata['QFILE-TEST-ASC.ASC']
        # read
        stream = _read_asc(testfile, skip=100, delta=0.1, length=2)
        stream.verify()
        # skip force one trace only
        assert len(stream) == 1
        # headers
        assert stream[0].stats.delta == 1.000000e-01
        assert stream[0].stats.npts == 2
        # check samples
        assert len(stream[0].data) == 2
        assert round(abs(stream[0].data[0]-111.7009), 4) == 0
        assert round(abs(stream[0].data[1]-119.5831), 4) == 0

    def test_write_small_trace(self):
        """
        Tests writing Traces containing 0, 1 or 2 samples only.
        """
        for format in ['SH_ASC', 'Q']:
            for num in range(0, 4):
                tr = Trace(data=np.arange(num))
                with NamedTemporaryFile() as tf:
                    tempfile = tf.name
                    if format == 'Q':
                        tempfile += '.QHD'
                    tr.write(tempfile, format=format)
                    # test results
                    with warnings.catch_warnings() as _:  # NOQA
                        warnings.simplefilter("ignore")
                        st = read(tempfile, format=format)
                    assert len(st) == 1
                    assert len(st[0]) == num
                    # Q files consist of two files - deleting additional file
                    if format == 'Q':
                        os.remove(tempfile[:-4] + '.QBN')
                        os.remove(tempfile[:-4] + '.QHD')

    def test_write_long_header(self):
        """
        Test for issue #1526
        """
        tr = read()[0]
        comment = 'This is a long comment for testing purposes.'
        tr.stats.sh = {'COMMENT': ' '.join(4 * [comment])}
        with NamedTemporaryFile(suffix='.QHD') as tf:
            tempfile = tf.name
            tr.write(tempfile, format="Q")
            tr2 = read(tempfile)[0]
            # remove binary file too (dynamically created)
            os.remove(os.path.splitext(tempfile)[0] + '.QBN')
        assert tr.stats.sh.COMMENT == tr2.stats.sh.COMMENT

    def test_header_whitespaces(self):
        """
        Test for issue #1552
        """
        tr = read()[0]
        tr.stats.sh = {'COMMENT': 30 * '   *   '}
        with NamedTemporaryFile(suffix='.QHD') as tf:
            tempfile = tf.name
            tr.write(tempfile, format="Q")
            tr2 = read(tempfile)[0]
            # remove binary file too (dynamically created)
            os.remove(os.path.splitext(tempfile)[0] + '.QBN')
        assert len(tr.stats.sh.COMMENT) == len(tr2.stats.sh.COMMENT)
        assert tr.stats.sh.COMMENT == tr2.stats.sh.COMMENT

    def test_append_traces(self):
        """
        Test for issue #2870
        """
        stream = read()
        with NamedTemporaryFile(suffix='.QHD') as tf:
            tempfile = tf.name
            stream.write(tempfile, format="Q")
            with open(tempfile) as f:
                header1 = f.read()
            # remove binary file too (dynamically created)
            os.remove(os.path.splitext(tempfile)[0] + '.QBN')
        with NamedTemporaryFile(suffix='.QHD') as tf:
            tempfile = tf.name
            stream[:1].write(tempfile, format="Q")
            stream[1:].write(tempfile, format="Q", append=True)
            with open(tempfile) as f:
                header2 = f.read()
            # remove binary file too (dynamically created)
            os.remove(os.path.splitext(tempfile)[0] + '.QBN')
        assert header2 == header1



# -*- coding: utf-8 -*-
# -------------------------------------------------------------------
# Filename: trigger.py
#  Purpose: Python trigger/picker routines for seismology.
#   Author: Moritz Beyreuther, Tobias Megies
#    Email: moritz.beyreuther@geophysik.uni-muenchen.de
#
# Copyright (C) 2008-2012 Moritz Beyreuther, Tobias Megies
# -------------------------------------------------------------------
"""
Various routines related to triggering/picking

Module implementing the Recursive STA/LTA. Two versions, a fast ctypes one and
a bit slower python one. Furthermore, the classic and delayed STA/LTA, the
carl_sta_trig and the z_detect are implemented.
Also includes picking routines, routines for evaluation and visualization of
characteristic functions and a coincidence triggering routine.

.. seealso:: [Withers1998]_ (p. 98) and [Trnkoczy2012]_

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
from collections import deque
import ctypes as C  # NOQA
import warnings

import numpy as np
import scipy

from obspy import UTCDateTime
from obspy.signal.cross_correlation import templates_max_similarity
from obspy.signal.headers import clibsignal, head_stalta_t


def recursive_sta_lta(a, nsta, nlta):
    """
    Recursive STA/LTA.

    Fast version written in C.

    :note: This version directly uses a C version via CTypes
    :type a: :class:`numpy.ndarray`, dtype=float64
    :param a: Seismic Trace, numpy.ndarray dtype float64
    :type nsta: int
    :param nsta: Length of short time average window in samples
    :type nlta: int
    :param nlta: Length of long time average window in samples
    :rtype: :class:`numpy.ndarray`, dtype=float64
    :return: Characteristic function of recursive STA/LTA

    .. seealso:: [Withers1998]_ (p. 98) and [Trnkoczy2012]_
    """
    # be nice and adapt type if necessary
    a = np.ascontiguousarray(a, np.float64)
    ndat = len(a)
    charfct = np.empty(ndat, dtype=np.float64)
    # do not use pointer here:
    clibsignal.recstalta(a, charfct, ndat, nsta, nlta)
    return charfct


def recursive_sta_lta_py(a, nsta, nlta):
    """
    Recursive STA/LTA written in Python.

    .. note::

        There exists a faster version of this trigger wrapped in C
        called :func:`~obspy.signal.trigger.recursive_sta_lta` in this module!

    :type a: NumPy :class:`~numpy.ndarray`
    :param a: Seismic Trace
    :type nsta: int
    :param nsta: Length of short time average window in samples
    :type nlta: int
    :param nlta: Length of long time average window in samples
    :rtype: NumPy :class:`~numpy.ndarray`
    :return: Characteristic function of recursive STA/LTA

    .. seealso:: [Withers1998]_ (p. 98) and [Trnkoczy2012]_
    """
    ndat = len(a)
    # compute the short time average (STA) and long time average (LTA)
    # given by Evans and Allen
    csta = 1. / nsta
    clta = 1. / nlta
    sta = 0.
    lta = np.finfo(0.0).tiny  # avoid zero division
    a = np.square(a)
    charfct = np.zeros(ndat, dtype=np.float64)
    icsta = 1 - csta
    iclta = 1 - clta
    for i in range(1, ndat):
        sta = csta * a[i] + icsta * sta
        lta = clta * a[i] + iclta * lta
        charfct[i] = sta / lta
    charfct[:nlta] = 0

    return charfct


def carl_sta_trig(a, nsta, nlta, ratio, quiet):
    """
    Computes the carlSTAtrig characteristic function.

    eta = star - (ratio * ltar) - abs(sta - lta) - quiet

    :type a: NumPy :class:`~numpy.ndarray`
    :param a: Seismic Trace
    :type nsta: int
    :param nsta: Length of short time average window in samples
    :type nlta: int
    :param nlta: Length of long time average window in samples
    :type ration: float
    :param ratio: as ratio gets smaller, carl_sta_trig gets more sensitive
    :type quiet: float
    :param quiet: as quiet gets smaller, carl_sta_trig gets more sensitive
    :rtype: NumPy :class:`~numpy.ndarray`
    :return: Characteristic function of CarlStaTrig
    """
    m = len(a)
    #
    sta = np.zeros(len(a), dtype=np.float64)
    lta = np.zeros(len(a), dtype=np.float64)
    star = np.zeros(len(a), dtype=np.float64)
    ltar = np.zeros(len(a), dtype=np.float64)
    pad_sta = np.zeros(nsta)
    pad_lta = np.zeros(nlta)  # avoid for 0 division 0/1=0
    #
    # compute the short time average (STA)
    for i in range(nsta):  # window size to smooth over
        sta += np.concatenate((pad_sta, a[i:m - nsta + i]))
    sta /= nsta
    #
    # compute the long time average (LTA), 8 sec average over sta
    for i in range(nlta):  # window size to smooth over
        lta += np.concatenate((pad_lta, sta[i:m - nlta + i]))
    lta /= nlta
    lta = np.concatenate((np.zeros(1), lta))[:m]  # XXX ???
    #
    # compute star, average of abs diff between trace and lta
    for i in range(nsta):  # window size to smooth over
        star += np.concatenate((pad_sta,
                               abs(a[i:m - nsta + i] - lta[i:m - nsta + i])))
    star /= nsta
    #
    # compute ltar, 8 sec average over star
    for i in range(nlta):  # window size to smooth over
        ltar += np.concatenate((pad_lta, star[i:m - nlta + i]))
    ltar /= nlta
    #
    eta = star - (ratio * ltar) - abs(sta - lta) - quiet
    eta[:nlta] = -1.0
    return eta


def classic_sta_lta(a, nsta, nlta):
    """
    Computes the standard STA/LTA from a given input array a. The length of
    the STA is given by nsta in samples, respectively is the length of the
    LTA given by nlta in samples.

    Fast version written in C.

    :type a: NumPy :class:`~numpy.ndarray`
    :param a: Seismic Trace
    :type nsta: int
    :param nsta: Length of short time average window in samples
    :type nlta: int
    :param nlta: Length of long time average window in samples
    :rtype: NumPy :class:`~numpy.ndarray`
    :return: Characteristic function of classic STA/LTA
    """
    data = a
    # initialize C struct / NumPy structured array
    head = np.empty(1, dtype=head_stalta_t)
    head[:] = (len(data), nsta, nlta)
    # ensure correct type and contiguous of data
    data = np.ascontiguousarray(data, dtype=np.float64)
    # all memory should be allocated by python
    charfct = np.empty(len(data), dtype=np.float64)
    # run and check the error-code
    errcode = clibsignal.stalta(head, data, charfct)
    if errcode != 0:
        raise Exception('ERROR %d stalta: len(data) < nlta' % errcode)
    return charfct


def classic_sta_lta_py(a, nsta, nlta):
    """
    Computes the standard STA/LTA from a given input array a. The length of
    the STA is given by nsta in samples, respectively is the length of the
    LTA given by nlta in samples. Written in Python.

    .. note::

        There exists a faster version of this trigger wrapped in C
        called :func:`~obspy.signal.trigger.classic_sta_lta` in this module!

    :type a: NumPy :class:`~numpy.ndarray`
    :param a: Seismic Trace
    :type nsta: int
    :param nsta: Length of short time average window in samples
    :type nlta: int
    :param nlta: Length of long time average window in samples
    :rtype: NumPy :class:`~numpy.ndarray`
    :return: Characteristic function of classic STA/LTA
    """
    # The cumulative sum can be exploited to calculate a moving average (the
    # cumsum function is quite efficient)
    sta = np.cumsum(a ** 2, dtype=np.float64)

    # Copy for LTA
    lta = sta.copy()

    # Compute the STA and the LTA
    sta[nsta:] = sta[nsta:] - sta[:-nsta]
    sta /= nsta
    lta[nlta:] = lta[nlta:] - lta[:-nlta]
    lta /= nlta

    # Pad zeros
    sta[:nlta - 1] = 0

    # Avoid division by zero by setting zero values to tiny float
    dtiny = np.finfo(0.0).tiny
    idx = lta < dtiny
    lta[idx] = dtiny

    return sta / lta


def delayed_sta_lta(a, nsta, nlta):
    """
    Delayed STA/LTA.

    :type a: NumPy :class:`~numpy.ndarray`
    :param a: Seismic Trace
    :type nsta: int
    :param nsta: Length of short time average window in samples
    :type nlta: int
    :param nlta: Length of long time average window in samples
    :rtype: NumPy :class:`~numpy.ndarray`
    :return: Characteristic function of delayed STA/LTA

    .. seealso:: [Withers1998]_ (p. 98) and [Trnkoczy2012]_
    """
    m = len(a)
    #
    # compute the short time average (STA) and long time average (LTA)
    # don't start for STA at nsta because it's muted later anyway
    sta = np.zeros(m, dtype=np.float64)
    lta = np.zeros(m, dtype=np.float64)
    for i in range(m):
        sta[i] = (a[i] ** 2 + a[i - nsta] ** 2) / nsta + sta[i - 1]
        lta[i] = (a[i - nsta - 1] ** 2 + a[i - nsta - nlta - 1] ** 2) / \
            nlta + lta[i - 1]
    sta[0:nlta + nsta + 50] = 0
    lta[0:nlta + nsta + 50] = 1  # avoid division by zero
    return sta / lta


def z_detect(a, nsta):
    """
    Z-detector.

    :param nsta: Window length in Samples.

    .. seealso:: [Withers1998]_, p. 99
    """
    # Z-detector given by Swindell and Snell (1977)
    # Standard Sta shifted by 1
    sta = np.cumsum(a ** 2, dtype=np.float64)
    sta[nsta + 1:] = sta[nsta:-1] - sta[:-nsta - 1]
    sta[nsta] = sta[nsta - 1]
    sta[:nsta] = 0
    a_mean = np.mean(sta)
    a_std = np.std(sta)
    _z = (sta - a_mean) / a_std
    return _z


def energy_ratio(a, nsta):
    r"""
    Energy ratio detector.

    Energy ratio defined as

    .. math::
        \text{er}(i) = \frac{\sum_{j=i}^{i+L}{a_j^2}}{\sum_{j=i-L}^{i}{a_j^2}}

    where :math:`L` is ``nsta``.

    :type a: NumPy :class:`~numpy.ndarray`
    :param a: Seismic Trace
    :type nsta: int
    :param nsta: Length of the energy ratio window in samples. It's the same
                 length as ``nsta`` in the classical STA/LTA methods.
    :rtype: NumPy :class:`~numpy.ndarray`
    :return: Energy Ratio

    .. seealso:: [Han2009]_
    """
    if nsta > len(a) // 2:
        # Half forward, half backward -> empty medium
        msg = (
            f'nsta ({nsta}) must not be larger than half the length of the '
            f'data ({len(a)} samples).')
        raise ValueError(msg)
    if nsta <= 0:
        # If nsta is zero, the sum is undefined
        msg = f'nsta ({nsta}) must not be equal to or less than zero.'
        raise ValueError(msg)
    sig_power = np.r_[0, np.cumsum(a ** 2, dtype=np.float64)]
    energy_diff = sig_power[nsta:] - sig_power[:len(sig_power) - nsta]
    er = np.zeros(len(a), dtype=np.float64)
    np.divide(energy_diff[nsta:], energy_diff[:len(energy_diff) - nsta],
              where=energy_diff[:len(energy_diff) - nsta] != 0,
              out=er[nsta:len(er) - nsta + 1])
    return er


def modified_energy_ratio(a, nsta, power=3):
    r"""
    Modified energy ratio detector.

    Improvement of the :func:`energy_ratio` that accounts for the signal
    itself:

    .. math::
        \text{mer}(i) = (\text{er}(i) * |a(i)|)^3

    where :math:`text{er}(i)` is the :func:`energy_ratio`.

    :type a: NumPy :class:`~numpy.ndarray`
    :param a: Seismic Trace
    :type nsta: int
    :param nsta: Length of energy ratio window in samples. It's the same length
                 as ``nsta`` in the classical STA/LTA methods.
    :type power: int
    :param power: The power exponent in the equation above. Default: 3
    :rtype: NumPy :class:`~numpy.ndarray`
    :return: Modified Energy Ratio
    """
    er = energy_ratio(a, nsta=nsta)
    mer = np.power(er * np.abs(a), power, out=er)
    return mer


def trigger_onset(charfct, thres1, thres2, max_len=9e99, max_len_delete=False):
    """
    Calculate trigger on and off times.

    Given thres1 and thres2 calculate trigger on and off times from
    characteristic function.

    This method is written in pure Python and gets slow as soon as there
    are more then 1e6 triggerings ("on" AND "off") in charfct --- normally
    this does not happen.

    :type charfct: NumPy :class:`~numpy.ndarray`
    :param charfct: Characteristic function of e.g. STA/LTA trigger
    :type thres1: float
    :param thres1: Value above which trigger (of characteristic function)
                   is activated (higher threshold)
    :type thres2: float
    :param thres2: Value below which trigger (of characteristic function)
        is deactivated (lower threshold)
    :type max_len: int
    :param max_len: Maximum length of triggered event in samples. A new
                    event will be triggered as soon as the signal reaches
                    again above thres1.
    :type max_len_delete: bool
    :param max_len_delete: Do not write events longer than max_len into
                           report file.
    :rtype: List
    :return: Nested List of trigger on and of times in samples
    """
    # 1) find indices of samples greater than threshold
    # 2) calculate trigger "of" times by the gap in trigger indices
    #    above the threshold i.e. the difference of two following indices
    #    in ind is greater than 1
    # 3) in principle the same as for "of" just add one to the index to get
    #    start times, this operation is not supported on the compact
    #    syntax
    # 4) as long as there is a on time greater than the actual of time find
    #    trigger on states which are greater than last of state an the
    #    corresponding of state which is greater than current on state
    # 5) if the signal stays above thres2 longer than max_len an event
    #    is triggered and following a new event can be triggered as soon as
    #    the signal is above thres1
    ind1 = np.where(charfct >= thres1)[0]
    if len(ind1) == 0:
        return []
    ind2 = np.where(charfct >= thres2)[0]
    #
    on = deque([ind1[0]])
    of = deque([-1])
    # determine the indices where charfct falls below off-threshold
    ind2_ = np.empty_like(ind2, dtype=bool)
    ind2_[:-1] = np.diff(ind2) > 1
    # last occurence is missed by the diff, add it manually
    ind2_[-1] = True
    of.extend(ind2[ind2_].tolist())
    on.extend(ind1[np.where(np.diff(ind1) > 1)[0] + 1].tolist())
    # include last pick if trigger is on or drop it
    if max_len_delete:
        # drop it
        of.extend([1e99])
        on.extend([on[-1]])
    else:
        # include it
        of.extend([ind2[-1]])

    # add last sample to ensure trigger gets switched off if ctf does not fall
    # below off-threshold before hitting the end
    of.append(len(charfct))
    #
    pick = []
    while on[-1] > of[0]:
        while on[0] <= of[0]:
            on.popleft()
        while of[0] < on[0]:
            of.popleft()
        if of[0] - on[0] > max_len:
            if max_len_delete:
                on.popleft()
                continue
            of.appendleft(on[0] + max_len)
        pick.append([on[0], of[0]])
    return np.array(pick, dtype=np.int64)


def pk_baer(reltrc, samp_int, tdownmax, tupevent, thr1, thr2, preset_len,
            p_dur, return_cf=False):
    """
    Wrapper for P-picker routine by M. Baer, Schweizer Erdbebendienst.

    :param reltrc: time series as numpy.ndarray float32 data, possibly filtered
    :param samp_int: number of samples per second
    :param tdownmax: if dtime exceeds tdownmax, the trigger is examined for
        validity
    :param tupevent: min nr of samples for itrm to be accepted as a pick
    :param thr1: threshold to trigger for pick (c.f. paper)
    :param thr2: threshold for updating sigma  (c.f. paper)
    :param preset_len: no of points taken for the estimation of variance of
        SF(t) on preset()
    :param p_dur: p_dur defines the time interval for which the maximum
        amplitude is evaluated Originally set to 6 secs
    :type return_cf: bool
    :param return_cf: If ``True``, also return the characteristic function.
    :return: (pptime, pfm [,cf]) pptime sample number of parrival;
        pfm direction of first motion (U or D), optionally also the
        characteristic function.

    .. note:: currently the first sample is not taken into account

    .. seealso:: [Baer1987]_
    """
    pptime = C.c_int()
    # c_chcar_p strings are immutable, use string_buffer for pointers
    pfm = C.create_string_buffer(b"     ", 5)
    # be nice and adapt type if necessary
    reltrc = np.ascontiguousarray(reltrc, np.float32)
    # Initiliaze CF array (MB)
    c_float_p = C.POINTER(C.c_float)
    cf_arr = np.zeros(len(reltrc) - 1, dtype=np.float32, order="C")
    cf_p = cf_arr.ctypes.data_as(c_float_p)
    # index in pk_mbaer.c starts with 1, 0 index is lost, length must be
    # one shorter
    args = (len(reltrc) - 1, C.byref(pptime), pfm, samp_int,
            tdownmax, tupevent, thr1, thr2, preset_len, p_dur, cf_p)
    errcode = clibsignal.ppick(reltrc, *args)
    if errcode != 0:
        raise MemoryError("Error in function ppick of mk_mbaer.c")
    # Switch cf_arr param (MB)
    # add the sample to the time which is not taken into account
    # pfm has to be decoded from byte to string
    if return_cf:
        return pptime.value + 1, pfm.value.decode('utf-8'), cf_arr
    else:
        return pptime.value + 1, pfm.value.decode('utf-8')


def aic_simple(a):
    r"""
    Simple Akaike Information Criterion [Maeda1985]_.

    It's computed directly from input data :math:`a` and defined as

    .. math::
        \text{AIC}(k) = k\log(\text{Var}(a_{1..k})) +
                        (N-k-1)\log(\text{Var}(a_{k+1..N}))

    which variance denoted as :math:`\text{Var}`.

    The true output is one data sample less. To make it convenient with other
    metrics in this module, where the output length is preserved, the last
    element is appended to the output: ``aic[-2] == aic[-1]``.

    :type a: :class:`numpy.ndarray` or :class:`list`
    :param a: Input time series
    :rtype: :class:`numpy.ndarray`
    :return: aic - Akaike Information Criterion array
    """
    n = len(a)
    if n <= 2:
        return np.zeros(n, dtype=np.float64)
    a = np.ascontiguousarray(a, np.float64)
    aic_res = np.empty(n, dtype=np.float64)
    clibsignal.aic_simple(aic_res, a, n)
    return aic_res


def ar_pick(a, b, c, samp_rate, f1, f2, lta_p, sta_p, lta_s, sta_s, m_p, m_s,
            l_p, l_s, s_pick=True):
    """
    Pick P and S arrivals with an AR-AIC + STA/LTA algorithm.

    The algorithm picks onset times using an Auto Regression - Akaike
    Information Criterion (AR-AIC) method. The detection intervals are
    successively narrowed down with the help of STA/LTA ratios as well as
    STA-LTA difference calculations. For details, please see [Akazawa2004]_.

    An important feature of this algorithm is that it requires comparatively
    little tweaking and site-specific settings and is thus applicable to large,
    diverse data sets.

    :type a: :class:`numpy.ndarray`
    :param a: Z signal the data.
    :type b: :class:`numpy.ndarray`
    :param b: N signal of the data.
    :type c: :class:`numpy.ndarray`
    :param c: E signal of the data.
    :type samp_rate: float
    :param samp_rate: Number of samples per second.
    :type f1: float
    :param f1: Frequency of the lower bandpass window.
    :type f2: float
    :param f2: Frequency of the upper .andpass window.
    :type lta_p: float
    :param lta_p: Length of LTA for the P arrival in seconds.
    :type sta_p: float
    :param sta_p: Length of STA for the P arrival in seconds.
    :type lta_s: float
    :param lta_s: Length of LTA for the S arrival in seconds.
    :type sta_s: float
    :param sta_s: Length of STA for the S arrival in seconds.
    :type m_p: int
    :param m_p: Number of AR coefficients for the P arrival.
    :type m_s: int
    :param m_s: Number of AR coefficients for the S arrival.
    :type l_p: float
    :param l_p: Length of variance window for the P arrival in seconds.
    :type l_s: float
    :param l_s: Length of variance window for the S arrival in seconds.
    :type s_pick: bool
    :param s_pick: If ``True``, also pick the S phase, otherwise only the P
        phase.
    :rtype: tuple
    :returns: A tuple with the P and the S arrival.
    """
    if not (len(a) == len(b) == len(c)):
        raise ValueError("All three data arrays must have the same length.")

    a = scipy.signal.detrend(a, type='linear')
    b = scipy.signal.detrend(b, type='linear')
    c = scipy.signal.detrend(c, type='linear')

    # be nice and adapt type if necessary
    a = np.require(a, dtype=np.float32, requirements=['C_CONTIGUOUS'])
    b = np.require(b, dtype=np.float32, requirements=['C_CONTIGUOUS'])
    c = np.require(c, dtype=np.float32, requirements=['C_CONTIGUOUS'])

    # scale amplitudes to avoid precision issues in case of low amplitudes
    # C code picks the horizontal component with larger amplitudes, so scale
    # horizontal components with a common scaling factor
    data_max = np.abs(a).max()
    if data_max < 100:
        a *= 1e6
        a /= data_max
    data_max = max(np.abs(b).max(), np.abs(c).max())
    if data_max < 100:
        for data in (b, c):
            data *= 1e6
            data /= data_max

    s_pick = C.c_int(s_pick)  # pick S phase also
    ptime = C.c_float()
    stime = C.c_float()
    args = (len(a), samp_rate, f1, f2,
            lta_p, sta_p, lta_s, sta_s, m_p, m_s, C.byref(ptime),
            C.byref(stime), l_p, l_s, s_pick)
    errcode = clibsignal.ar_picker(a, b, c, *args)
    if errcode != 0:
        bufs = ['buff1', 'buff1_s', 'buff2', 'buff3', 'buff4', 'buff4_s',
                'f_error', 'b_error', 'ar_f', 'ar_b', 'buf_sta', 'buf_lta',
                'extra_tr1', 'extra_tr2', 'extra_tr3']
        if errcode <= len(bufs):
            raise MemoryError('Unable to allocate %s!' % (bufs[errcode - 1]))
        raise Exception('Error during PAZ calculation!')
    return ptime.value, stime.value


def plot_trace(trace, cft):
    """
    Plot characteristic function of trigger along with waveform data.

    :type trace: :class:`~obspy.core.trace.Trace`
    :param trace: waveform data
    :type cft: :class:`numpy.ndarray`
    :param cft: characteristic function as returned by a trigger in
        :mod:`obspy.signal.trigger`
    :rtype: tuple
    :returns: Matplotlib figure instance and axes
    """
    import matplotlib.pyplot as plt
    df = trace.stats.sampling_rate
    npts = trace.stats.npts
    t = np.arange(npts, dtype=np.float32) / df
    fig, axes = plt.subplots(nrows=2, sharex=True)
    axes[0].plot(t, trace.data, 'k')
    axes[1].plot(t, cft, 'k')
    axes[1].set_xlabel("Time after %s [s]" % trace.stats.starttime.isoformat())
    fig.suptitle(trace.id)
    fig.canvas.draw()
    return fig, axes


def plot_trigger(trace, cft, thr_on, thr_off, show=True):
    """
    Plot characteristic function of trigger along with waveform data and
    trigger On/Off from given thresholds.

    :type trace: :class:`~obspy.core.trace.Trace`
    :param trace: waveform data
    :type cft: :class:`numpy.ndarray`
    :param cft: characteristic function as returned by a trigger in
        :mod:`obspy.signal.trigger`
    :type thr_on: float
    :param thr_on: threshold for switching trigger on
    :type thr_off: float
    :param thr_off: threshold for switching trigger off
    :type show: bool
    :param show: Do not call `plt.show()` at end of routine. That way,
        further modifications can be done to the figure before showing it.
    :rtype: tuple
    :returns: Matplotlib figure instance and axes
    """
    import matplotlib.pyplot as plt
    df = trace.stats.sampling_rate
    fig, axes = plot_trace(trace, cft)
    ax1, ax2 = axes
    on_off = np.array(trigger_onset(cft, thr_on, thr_off))
    ymin, ymax = ax1.get_ylim()
    try:
        ax1.vlines(on_off[:, 0] / df, ymin, ymax, color='r', lw=2,
                   label="Trigger On")
        ax1.vlines(on_off[:, 1] / df, ymin, ymax, color='b', lw=2,
                   label="Trigger Off")
        ax1.legend()
    except IndexError:
        pass
    ax2.axhline(thr_on, color='red', lw=1, ls='--')
    ax2.axhline(thr_off, color='blue', lw=1, ls='--')
    ax2.set_xlabel("Time after %s [s]" % trace.stats.starttime.isoformat())
    fig.suptitle(trace.id)
    fig.canvas.draw()
    if show:
        plt.show()
    return fig, axes


def coincidence_trigger(trigger_type, thr_on, thr_off, stream,
                        thr_coincidence_sum, trace_ids=None,
                        max_trigger_length=1e6, delete_long_trigger=False,
                        trigger_off_extension=0, details=False,
                        event_templates={}, similarity_threshold=0.7,
                        **options):
    """
    Perform a network coincidence trigger.

    The routine works in the following steps:
      * take every single trace in the stream
      * apply specified triggering routine (can be skipped to work on
        precomputed custom characteristic functions)
      * evaluate all single station triggering results
      * compile chronological overall list of all single station triggers
      * find overlapping single station triggers
      * calculate coincidence sum of every individual overlapping trigger
      * add to coincidence trigger list if it exceeds the given threshold
      * optional: if master event templates are provided, also check single
        station triggers individually and include any single station trigger if
        it exceeds the specified similarity threshold even if no other stations
        coincide with the trigger
      * return list of network coincidence triggers

    .. note::
        An example can be found in the
        `Trigger/Picker Tutorial
        <https://tutorial.obspy.org/code_snippets/trigger_tutorial.html>`_.

    .. note::
        Setting `trigger_type=None` precomputed characteristic functions can
        be provided.

    .. seealso:: [Withers1998]_ (p. 98) and [Trnkoczy2012]_

    :param trigger_type: String that specifies which trigger is applied (e.g.
        ``'recstalta'``). See e.g. :meth:`obspy.core.trace.Trace.trigger` for
        further details. If set to `None` no triggering routine is applied,
        i.e.  data in traces is supposed to be a precomputed characteristic
        function on which the trigger thresholds are evaluated.
    :type trigger_type: str or None
    :type thr_on: float
    :param thr_on: threshold for switching single station trigger on
    :type thr_off: float
    :param thr_off: threshold for switching single station trigger off
    :type stream: :class:`~obspy.core.stream.Stream`
    :param stream: Stream containing waveform data for all stations. These
        data are changed inplace, make a copy to keep the raw waveform data.
    :type thr_coincidence_sum: int or float
    :param thr_coincidence_sum: Threshold for coincidence sum. The network
        coincidence sum has to be at least equal to this value for a trigger to
        be included in the returned trigger list.
    :type trace_ids: list or dict, optional
    :param trace_ids: Trace IDs to be used in the network coincidence sum. A
        dictionary with trace IDs as keys and weights as values can
        be provided. If a list of trace IDs is provided, all
        weights are set to 1. The default of ``None`` uses all traces present
        in the provided stream. Waveform data with trace IDs not
        present in this list/dict are disregarded in the analysis.
    :type max_trigger_length: int or float
    :param max_trigger_length: Maximum single station trigger length (in
        seconds). ``delete_long_trigger`` controls what happens to single
        station triggers longer than this value.
    :type delete_long_trigger: bool, optional
    :param delete_long_trigger: If ``False`` (default), single station
        triggers are manually released at ``max_trigger_length``, although the
        characteristic function has not dropped below ``thr_off``. If set to
        ``True``, all single station triggers longer than
        ``max_trigger_length`` will be removed and are excluded from
        coincidence sum computation.
    :type trigger_off_extension: int or float, optional
    :param trigger_off_extension: Extends search window for next trigger
        on-time after last trigger off-time in coincidence sum computation.
    :type details: bool, optional
    :param details: If set to ``True`` the output coincidence triggers contain
        more detailed information: A list with the trace IDs (in addition to
        only the station names), as well as lists with single station
        characteristic function peak values and standard deviations in the
        triggering interval and mean values of both, relatively weighted like
        in the coincidence sum. These values can help to judge the reliability
        of the trigger.
    :param options: Necessary keyword arguments for the respective trigger
        that will be passed on. For example ``sta`` and ``lta`` for any STA/LTA
        variant (e.g. ``sta=3``, ``lta=10``).
        Arguments ``sta`` and ``lta`` (seconds) will be mapped to ``nsta``
        and ``nlta`` (samples) by multiplying with sampling rate of trace.
        (e.g. ``sta=3``, ``lta=10`` would call the trigger with 3 and 10
        seconds average, respectively)
    :param event_templates: Event templates to use in checking similarity of
        single station triggers against known events. Expected are streams with
        three traces for Z, N, E component. A dictionary is expected where for
        each station used in the trigger, a list of streams can be provided as
        the value to the network/station key (e.g. {"GR.FUR": [stream1,
        stream2]}). Templates are compared against the provided `stream`
        without the specified triggering routine (`trigger_type`) applied.
    :type event_templates: dict
    :param similarity_threshold: similarity threshold (0.0-1.0) at which a
        single station trigger gets included in the output network event
        trigger list. A common threshold can be set for all stations (float) or
        a dictionary mapping station names to float values for each station.
    :type similarity_threshold: float or dict
    :rtype: list
    :returns: List of event triggers sorted chronologically.
    """
    # if no trace ids are specified use all traces ids found in stream
    if trace_ids is None:
        trace_ids = [tr.id for tr in stream]
    # we always work with a dictionary with trace ids and their weights later
    if isinstance(trace_ids, list) or isinstance(trace_ids, tuple):
        trace_ids = dict.fromkeys(trace_ids, 1)
    # set up similarity thresholds as a dictionary if necessary
    if not isinstance(similarity_threshold, dict):
        similarity_threshold = dict.fromkeys(
            [tr.stats.station for tr in stream], similarity_threshold)

    # the single station triggering
    triggers = []
    # prepare kwargs for trigger_onset
    kwargs = {'max_len_delete': delete_long_trigger}
    for tr in stream:
        tr = tr.copy()
        if tr.id not in trace_ids:
            msg = "At least one trace's ID was not found in the " + \
                  "trace ID list and was disregarded (%s)" % tr.id
            warnings.warn(msg, UserWarning)
            continue
        if trigger_type is not None:
            tr.trigger(trigger_type, **options)
        kwargs['max_len'] = int(
            max_trigger_length * tr.stats.sampling_rate + 0.5)
        tmp_triggers = trigger_onset(tr.data, thr_on, thr_off, **kwargs)
        for on, off in tmp_triggers:
            try:
                cft_peak = tr.data[on:off].max()
                cft_std = tr.data[on:off].std()
            except ValueError:
                cft_peak = tr.data[on]
                cft_std = 0
            on = tr.stats.starttime + float(on) / tr.stats.sampling_rate
            off = tr.stats.starttime + float(off) / tr.stats.sampling_rate
            triggers.append((on.timestamp, off.timestamp, tr.id, cft_peak,
                             cft_std))
    triggers.sort()

    for i, (on, off, tr_id, cft_peak, cft_std) in enumerate(triggers):
        sta = tr_id.split(".")[1]
        templates = event_templates.get(sta)
        if templates:
            simil = templates_max_similarity(
                stream, UTCDateTime(on), templates)
        else:
            simil = None
        triggers[i] = (on, off, tr_id, cft_peak, cft_std, simil)

    # the coincidence triggering and coincidence sum computation
    coincidence_triggers = []
    last_off_time = 0.0
    while triggers != []:
        # remove first trigger from list and look for overlaps
        on, off, tr_id, cft_peak, cft_std, simil = triggers.pop(0)
        sta = tr_id.split(".")[1]
        event = {}
        event['time'] = UTCDateTime(on)
        event['stations'] = [tr_id.split(".")[1]]
        event['trace_ids'] = [tr_id]
        event['coincidence_sum'] = float(trace_ids[tr_id])
        event['similarity'] = {}
        if details:
            event['cft_peaks'] = [cft_peak]
            event['cft_stds'] = [cft_std]
        # evaluate maximum similarity for station if event templates were
        # provided
        if simil is not None:
            event['similarity'][sta] = simil
        # compile the list of stations that overlap with the current trigger
        for (tmp_on, tmp_off, tmp_tr_id, tmp_cft_peak, tmp_cft_std,
                tmp_simil) in triggers:
            tmp_sta = tmp_tr_id.split(".")[1]
            # skip retriggering of already present station in current
            # coincidence trigger
            if tmp_tr_id in event['trace_ids']:
                continue
            # check for overlapping trigger,
            # break if there is a gap in between the two triggers
            if tmp_on > off + trigger_off_extension:
                break
            event['stations'].append(tmp_sta)
            event['trace_ids'].append(tmp_tr_id)
            event['coincidence_sum'] += trace_ids[tmp_tr_id]
            if details:
                event['cft_peaks'].append(tmp_cft_peak)
                event['cft_stds'].append(tmp_cft_std)
            # allow sets of triggers that overlap only on subsets of all
            # stations (e.g. A overlaps with B and B overlaps w/ C => ABC)
            off = max(off, tmp_off)
            # evaluate maximum similarity for station if event templates were
            # provided
            if tmp_simil is not None:
                event['similarity'][tmp_sta] = tmp_simil
        # skip if both coincidence sum and similarity thresholds are not met
        if event['coincidence_sum'] < thr_coincidence_sum:
            if not event['similarity']:
                continue
            elif not any([val > similarity_threshold[_s]
                          for _s, val in event['similarity'].items()]):
                continue
        # skip coincidence trigger if it is just a subset of the previous
        # (determined by a shared off-time, this is a bit sloppy)
        if off <= last_off_time:
            continue
        event['duration'] = off - on
        if details:
            weights = np.array([trace_ids[i] for i in event['trace_ids']])
            weighted_values = np.array(event['cft_peaks']) * weights
            event['cft_peak_wmean'] = weighted_values.sum() / weights.sum()
            weighted_values = np.array(event['cft_stds']) * weights
            event['cft_std_wmean'] = \
                (np.array(event['cft_stds']) * weights).sum() / weights.sum()
        coincidence_triggers.append(event)
        last_off_time = off
    return coincidence_triggers



# -*- coding: utf-8 -*-
SIZE_ALSEP_PSE_RECORD = 19456
SIZE_ALSEP_RECORD_HEADER = 16
SIZE_ALSEP_PSE_FRAME_OLD = 72
SIZE_ALSEP_PSE_FRAME_NEW = 36

FORMAT_ALSEP_PSE_OLD = 0
FORMAT_ALSEP_PSE_NEW = 1
FORMAT_ALSEP_WTN = 2
FORMAT_ALSEP_WTH = 3

DATA_RATE_530_BPS = 0
DATA_RATE_1060_BPS = 1

channels = {
    # The channel names should be determined as below:
    #
    # https://ds.iris.edu/ds/nodes/dmc/data/formats/seed-channel-naming/
    # https://www.geonet.org.nz/data/supplementary/channels
    # 'spz': {'channel': 'SHZ', 'sampling_rate': 53.0},
    # 'lpx': {'channel': 'MH1', 'sampling_rate': 6.625},
    # 'lpy': {'channel': 'MH2', 'sampling_rate': 6.625},
    # 'lpz': {'channel': 'MH3', 'sampling_rate': 6.625},
    # 'tdx': {'channel': 'LH1', 'sampling_rate': 0.828125},
    # 'tdy': {'channel': 'LH2', 'sampling_rate': 0.828125},
    # 'tdz': {'channel': 'LH3', 'sampling_rate': 0.828125},
    # 'inst_temp': {'channel': 'LK2', 'sampling_rate': 0.828125},
    # 'lsm': {'channel': 'FMF', 'sampling_rate': 9.9375},
    # 'lsm_status': {'channel': 'YLF', 'sampling_rate': 1.65625},
    # 'lsg': {'channel': 'SPZ', 'sampling_rate': 53.0},
    # ...
    #
    # Defined ALSEP specific channel names for clarity
    'spz': {'channel': 'SPZ', 'sampling_rate': 53.0},
    'lpx': {'channel': 'LPX', 'sampling_rate': 6.625},
    'lpy': {'channel': 'LPY', 'sampling_rate': 6.625},
    'lpz': {'channel': 'LPZ', 'sampling_rate': 6.625},
    'tdx': {'channel': 'TDX', 'sampling_rate': 0.828125},
    'tdy': {'channel': 'TDY', 'sampling_rate': 0.828125},
    'tdz': {'channel': 'TDZ', 'sampling_rate': 0.828125},
    'inst_temp': {'channel': 'TMP', 'sampling_rate': 0.828125},
    'lsm': {'channel': 'LSM', 'sampling_rate': 9.9375},
    'lsm_status': {'channel': 'LSS', 'sampling_rate': 1.65625},
    'lsg': {'channel': 'LSG', 'sampling_rate': 53.0},
    'lsg_tide': {'channel': 'LGT', 'sampling_rate': 1.65625},
    'lsg_free': {'channel': 'LGF', 'sampling_rate': 1.65625},
    'lsg_temp': {'channel': 'LGP', 'sampling_rate': 1.65625},
    'hk': {'channel': 'HK_', 'sampling_rate': 1.65625},
    'cv': {'channel': 'CV_', 'sampling_rate': 1.65625},
    # LUNAR SEISMIC PROFILING (LSP) EXPERIMENT
    #   3.5333 KBPS normal rate, NRZ-C PCM
    #   1.06 KBPS low rate, NRZ-C PCM
    #   1.963 frames per second normal rate
    #   0.588 frames per second low rate
    #   3 subframes per frame
    #   20 words per subframe
    #   30 bits per word
    # sampling rate is as below:
    #   20 data samples per frame in each channel.
    #   For normal rate:
    #   20 [samples/frame] x 1.963 [frames/second] = 39.26 [samples/second]
    #   For low rate:
    #   20 [samples/frame] x 0.588 [frames/second] = 11.76 [samples/second]
    'geo1': {'channel': 'GP1', 'sampling_rate': 39.26},
    'geo2': {'channel': 'GP2', 'sampling_rate': 39.26},
    'geo3': {'channel': 'GP3', 'sampling_rate': 39.26},
    'geo4': {'channel': 'GP4', 'sampling_rate': 39.26},
}

package_id_to_apollo_station = {
    1: 12,
    2: 15,
    3: 16,
    4: 14,
    5: 17,
}
